"""`ReferenceChecker`."""

import logging

from suricata_check.checkers.interface import CheckerInterface
from suricata_check.utils.checker import (
    is_rule_option_equal_to_regex,
)
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue, Rule
from suricata_check.utils.regex import get_regex_provider

_regex_provider = get_regex_provider()

_S700_REGEX = _regex_provider.compile(
    r"^(?!url).*[A-Z]+.*$",
)
_S701_REGEX = _regex_provider.compile(
    r"^url,\s*https?.*$",
    _regex_provider.IGNORECASE,
)


class ReferenceChecker(CheckerInterface):
    """The `ReferenceChecker` contains several checks for Suricata reference option.

    Codes S700-710 report on non-standard usages of `reference`
    """

    codes = {
        "S700": {"severity": logging.INFO},
        "S701": {"severity": logging.INFO},
    }

    def _check_rule(
        self: "ReferenceChecker",
        rule: Rule,
    ) -> ISSUES_TYPE:
        issues: ISSUES_TYPE = []

        if is_rule_option_equal_to_regex(rule, "reference", _S700_REGEX):
            issues.append(
                Issue(
                    code="S700",
                    message="""\
The rule uses uppercase characters in the `reference` option.
Consider using only lowercase characters.\
""",
                ),
            )

        if is_rule_option_equal_to_regex(rule, "reference", _S701_REGEX):
            issues.append(
                Issue(
                    code="S701",
                    message="""\
The rule specifies the web protocol in the `reference` option.
Consider removing the protocol.\
""",
                ),
            )

        return issues



import logging
import os

import click

_logger = logging.getLogger(__name__)


def find_rules_file(root: str) -> str:
    """Find the Suricata rules file in the given directory.

    Returns an absolute path to the rules file.
    """
    if not os.path.exists(root):
        msg = f"Error: {root} does not exist."
        _logger.critical(msg)
        raise click.BadParameter(f"Error: {msg}")

    is_root_dir = os.path.isdir(root)
    if not root.endswith(".rules") and not is_root_dir:
        msg = f"Error: {root} is not a rules file or directory."
        _logger.critical(msg)
        raise click.BadParameter(f"Error: {msg}")

    if not is_root_dir:
        rules_file = root
    else:
        full_path = os.path.abspath(root)
        _logger.info("Searching for Suricata rules file in %s", full_path)

        rules_files: list[str] = []
        for path, _, files in os.walk(full_path):
            for file in files:
                if file.endswith(".rules"):
                    rules_files.append(os.path.join(path, file))

        if len(rules_files) == 0:
            msg = f"No Suricata rules file found in {full_path}"
            _logger.critical(msg)
            raise click.BadParameter(f"Error: {msg}")
        if len(rules_files) > 1:
            msg = f"Multiple Suricata rules files found in {full_path}\n" + "\n".join(
                rules_files,
            )
            _logger.critical(msg)
            raise click.BadParameter(f"Error: {msg}")

        rules_file = rules_files[0]

    _logger.info("Found Suricata rules file: %s", rules_file)

    return rules_file



"""Main entrypoint for module referncing which can be used as a proxy for `suricata_check.suricata_check:main`."""

from suricata_check.suricata_check import main

if __name__ == "__main__":
    main()



"""The `suricata_check.suricata_check` module contains the command line utility and the main program logic."""

import atexit
import configparser
import io
import json
import logging
import logging.handlers
import multiprocessing
import os
import sys
import threading
from collections.abc import Sequence
from typing import (
    Any,
    Callable,
    Literal,
    Optional,
    TypeVar,
    Union,
)

import click

import suricata_check
from suricata_check.utils.checker_typing import Rule

_AnyCallable = Callable[..., Any]
_FC = TypeVar("_FC", bound=Union[_AnyCallable, click.Command])

LOG_LEVELS = ("DEBUG", "INFO", "WARNING", "ERROR")

# Add suricata-check to the front of the PATH, such that the version corresponding to the CLI is used.
_suricata_check_path = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
if sys.path[0] != _suricata_check_path:
    sys.path.insert(0, _suricata_check_path)

from suricata_check import (  # noqa: E402
    __version__,
    check_for_update,
    get_dependency_versions,
)
from suricata_check._checkers import get_checkers  # noqa: E402
from suricata_check._output import (  # noqa: E402
    summarize_output,
    summarize_rule,
    write_output,
)
from suricata_check.checkers.interface import CheckerInterface  # noqa: E402
from suricata_check.utils._click import ClickHandler, help_option  # noqa: E402
from suricata_check.utils._path import find_rules_file  # noqa: E402
from suricata_check.utils.checker import (  # noqa: E402
    check_rule_option_recognition,
    get_rule_suboption,
)
from suricata_check.utils.checker_typing import (  # noqa: E402
    InvalidRuleError,
    OutputReport,
    RuleReport,
)
from suricata_check.utils.regex import get_regex_provider, is_valid_rule  # noqa: E402

LOG_LEVELS = ("DEBUG", "INFO", "WARNING", "ERROR")
LogLevel = Literal["DEBUG", "INFO", "WARNING", "ERROR"]

# Define all command line arguments and their properties
CLI_ARGUMENTS: dict[str, dict[str, Any]] = {
    "rules": {
        "help": "Path to Suricata rules to provide check on.",
        "show_default": True,
        "type": str,
        "required": False,
        "default": ".",
        "cli_options": ["-r"],
    },
    "single_rule": {
        "help": "A single Suricata rule to be checked",
        "show_default": False,
        "type": str,
        "required": False,
        "default": None,
        "cli_options": ["-s"],
    },
    "out": {
        "help": "Path to suricata-check output folder.",
        "show_default": True,
        "type": str,
        "required": False,
        "default": ".",
        "cli_options": ["-o"],
    },
    "log_level": {
        "help": f"Verbosity level for logging. Can be one of {LOG_LEVELS}",
        "show_default": True,
        "type": str,
        "required": False,
        "default": "DEBUG",
    },
    "gitlab": {
        "help": "Flag to create CodeClimate output report for GitLab CI/CD.",
        "show_default": True,
        "type": bool,
        "required": False,
        "default": False,
        "is_flag": True,
    },
    "github": {
        "help": "Flag to write workflow commands to stdout for GitHub CI/CD.",
        "show_default": True,
        "type": bool,
        "required": False,
        "default": False,
        "is_flag": True,
    },
    "evaluate_disabled": {
        "help": "Flag to evaluate disabled rules.",
        "show_default": True,
        "type": bool,
        "required": False,
        "default": False,
        "is_flag": True,
    },
    "issue_severity": {
        "help": f"Verbosity level for detected issues. Can be one of {LOG_LEVELS}",
        "show_default": True,
        "type": str,
        "required": False,
        "default": "INFO",
    },
    "include_all": {
        "help": "Flag to indicate all checker codes should be enabled.",
        "show_default": True,
        "type": bool,
        "required": False,
        "default": False,
        "is_flag": True,
        "cli_options": ["-a"],
    },
    "include": {
        "help": "List of all checker codes to enable.",
        "show_default": True,
        "type": tuple,
        "required": False,
        "default": (),
        "multiple": True,
        "cli_options": ["-i"],
    },
    "exclude": {
        "help": "List of all checker codes to disable.",
        "show_default": True,
        "type": tuple,
        "required": False,
        "default": (),
        "multiple": True,
        "cli_options": ["-e"],
    },
}
CLI_ARGUMENT_TYPE = Optional[Union[str, bool, tuple]]

_logger = logging.getLogger(__name__)

_regex_provider = get_regex_provider()


def __create_click_option(name: str, props: dict[str, Any]) -> Callable[[_FC], _FC]:
    """Create a click.option decorator from argument properties."""
    kwargs = {
        "help": props["help"],
        "show_default": props["show_default"],
        "default": props.get("default"),
        "is_flag": props.get("is_flag", False),
        "multiple": props.get("multiple", False),
    }

    # Add any additional CLI options (like -r, -s, etc.)
    cli_opts = props.get("cli_options", [])
    args = [f"--{name.replace('_', '-')}", *cli_opts]

    return click.option(*args, **kwargs)


def __main_decorators() -> Callable[[Callable], click.Command]:
    """Create the CLI command with all options."""

    def decorator(f: Callable) -> click.Command:
        # Apply all options in reverse order (bottom to top)
        command = click.command()(f)
        command = help_option("-h", "--help")(command)

        # Add ini option first since it's needed before processing other options
        command = click.option(
            "--ini",
            help="Path to suricata-check.ini file to read configuration from.",
            show_default=True,
            type=str,
            default=None,
        )(command)

        # Apply options from CLI_ARGUMENTS
        for name, props in reversed(CLI_ARGUMENTS.items()):
            command = __create_click_option(name, props)(command)

        return command

    return decorator


@__main_decorators()
def main(**kwargs: dict[str, Any]) -> None:  # noqa: C901, PLR0915
    """The `suricata-check` command processes all rules inside a rules file and outputs a list of detected issues.

    Raises:
      BadParameter: If provided arguments are invalid.

      RuntimeError: If no checkers could be automatically discovered.

    """
    # Look for a ini file and parse it.
    ini_kwargs = get_ini_kwargs(
        str(kwargs["ini"]) if kwargs["ini"] is not None else None  # type: ignore reportUnnecessaryComparison
    )

    # Verify CLI argument types and get CLI arguments or use default arguments
    rules: str = __get_verified_kwarg(
        [kwargs, ini_kwargs], "rules"
    )  # pyright: ignore[reportAssignmentType]
    single_rule: Optional[str] = __get_verified_kwarg(
        [kwargs, ini_kwargs], "single_rule"
    )  # pyright: ignore[reportAssignmentType]
    out: str = __get_verified_kwarg(
        [kwargs, ini_kwargs], "out"
    )  # pyright: ignore[reportAssignmentType]
    log_level: LogLevel = __get_verified_kwarg(
        [kwargs, ini_kwargs], "log_level"
    )  # pyright: ignore[reportAssignmentType]
    gitlab: bool = __get_verified_kwarg(
        [kwargs, ini_kwargs], "gitlab"
    )  # pyright: ignore[reportAssignmentType]
    github: bool = __get_verified_kwarg(
        [kwargs, ini_kwargs], "github"
    )  # pyright: ignore[reportAssignmentType]
    evaluate_disabled: bool = __get_verified_kwarg(
        [kwargs, ini_kwargs], "evaluate_disabled"
    )  # pyright: ignore[reportAssignmentType]
    issue_severity: LogLevel = __get_verified_kwarg(
        [kwargs, ini_kwargs], "issue_severity"
    )  # pyright: ignore[reportAssignmentType]
    include_all: bool = __get_verified_kwarg(
        [kwargs, ini_kwargs], "include_all"
    )  # pyright: ignore[reportAssignmentType]
    include: tuple[str, ...] = __get_verified_kwarg(
        [kwargs, ini_kwargs], "include"
    )  # pyright: ignore[reportAssignmentType]
    exclude: tuple[str, ...] = __get_verified_kwarg(
        [kwargs, ini_kwargs], "exclude"
    )  # pyright: ignore[reportAssignmentType]

    # Verify that out argument is valid
    if os.path.exists(out) and not os.path.isdir(out):
        raise click.BadParameter(f"Error: {out} is not a directory.")

    # Verify that log_level argument is valid
    if log_level not in LOG_LEVELS:
        raise click.BadParameter(f"Error: {log_level} is not a valid log level.")

    # Create out directory if non-existent
    if not os.path.exists(out):
        os.makedirs(out)

    # Setup logging from a seperate thread
    queue = multiprocessing.Manager().Queue()
    queue_handler = logging.handlers.QueueHandler(queue)

    click_handler = ClickHandler(
        github=github, github_level=getattr(logging, log_level)
    )
    logging.basicConfig(
        level=log_level,
        format="%(asctime)s - %(name)s - %(levelname)s - %(threadName)s - %(message)s",
        handlers=(queue_handler, click_handler),
        force=os.environ.get("SURICATA_CHECK_FORCE_LOGGING", "FALSE") == "TRUE",
    )

    file_handler = logging.FileHandler(
        filename=os.path.join(out, "suricata-check.log"),
        delay=True,
    )
    queue_listener = logging.handlers.QueueListener(
        queue,
        file_handler,
        respect_handler_level=True,
    )

    def _at_exit() -> None:
        """Cleans up logging listener and handlers before exiting."""
        queue_listener.enqueue_sentinel()
        queue_listener.stop()
        file_handler.flush()
        file_handler.close()
        atexit.unregister(_at_exit)

    atexit.register(_at_exit)

    queue_listener.start()

    # Log the arguments:
    _logger.info("Running suricata-check with the following arguments:")
    for arg in CLI_ARGUMENTS:
        _logger.info("%s: %s", arg, locals().get(arg))

    # Log the environment:
    _logger.debug("Platform: %s", sys.platform)
    _logger.debug("Python version: %s", sys.version)
    _logger.debug("suricata-check path: %s", _suricata_check_path)
    _logger.debug("suricata-check version: %s", __version__)
    for package, version in get_dependency_versions().items():
        _logger.debug("Dependency %s version: %s", package, version)

    threading.Thread(
        target=check_for_update,
    ).start()

    # Verify that include and exclude arguments are valid
    if include_all and len(include) > 0:
        raise click.BadParameter(
            "Error: Cannot use --include-all and --include together."
        )
    if include_all:
        include = (".*",)

    # Verify that issue_severity argument is valid
    if issue_severity not in LOG_LEVELS:
        raise click.BadParameter(
            f"Error: {issue_severity} is not a valid issue severity or log level."
        )

    checkers = get_checkers(
        include, exclude, issue_severity=getattr(logging, issue_severity)
    )

    if single_rule is not None:
        __main_single_rule(out, single_rule, checkers)

        # Return here so no rules file is processed.
        _at_exit()
        return

    # Check if the rules argument is valid and find the rules file
    rules = find_rules_file(rules)

    output = process_rules_file(rules, evaluate_disabled, checkers=checkers)

    write_output(output, out, gitlab=gitlab, github=github, rules_file=rules)

    _at_exit()


def get_ini_kwargs(path: Optional[str]) -> dict[str, Any]:
    """Read configuration from INI file based on CLI_ARGUMENTS structure."""
    ini_kwargs: dict[str, Any] = {}
    if path is not None and not os.path.exists(path):
        raise click.BadParameter(
            f"Error: INI file provided in {path} but no options loaded"
        )

    # Use the default path if no path was provided
    if path is None:
        path = "suricata-check.ini"
        if not os.path.exists(path):
            return {}

    config_parser = configparser.ConfigParser(
        empty_lines_in_values=False,
        default_section="suricata-check",
        converters={"tuple": lambda x: tuple(json.loads(x))},
    )
    config_parser.read(path)

    # Process each argument defined in CLI_ARGUMENTS
    for arg_name, arg_props in CLI_ARGUMENTS.items():
        ini_key = arg_name.replace("_", "-")
        if not config_parser.has_option("suricata-check", ini_key):
            continue

        # Get the value based on the argument type
        if arg_props["type"] is bool:
            ini_kwargs[arg_name] = config_parser.getboolean("suricata-check", ini_key)
        elif arg_props["type"] is tuple:
            ini_kwargs[arg_name] = config_parser.gettuple("suricata-check", ini_key)  # type: ignore reportAttributeAccessIssue
        else:
            ini_kwargs[arg_name] = config_parser.get("suricata-check", ini_key)
            if arg_props["type"] is str:
                ini_kwargs[arg_name] = ini_kwargs[arg_name].strip('"')

    return ini_kwargs


def __get_verified_kwarg(
    kwargss: Sequence[dict[str, Any]],
    name: str,
) -> CLI_ARGUMENT_TYPE:
    for kwargs in kwargss:
        if name in kwargs:
            if kwargs[name] is None:
                if (
                    not CLI_ARGUMENTS[name]["required"]
                    and CLI_ARGUMENTS[name]["default"] is not None
                ):
                    return None
                return CLI_ARGUMENTS[name]["default"]

            if kwargs[name] is not CLI_ARGUMENTS[name]["default"]:
                if not isinstance(kwargs[name], CLI_ARGUMENTS[name]["type"]):
                    raise click.BadParameter(
                        f"""Error: \
                Argument `{name}` should have a value of type `{CLI_ARGUMENTS[name]["type"]}` \
                but has value {kwargs[name]} of type {kwargs[name].__class__} instead."""
                    )
                return kwargs[name]

    return CLI_ARGUMENTS[name]["default"]


def __main_single_rule(
    out: str, single_rule: str, checkers: Optional[Sequence[CheckerInterface]]
) -> None:
    rule: Optional[Rule] = suricata_check.rule.parse(single_rule)

    # Verify that a rule was parsed correctly.
    if rule is None:
        msg = f"Error parsing rule from user input: {single_rule}"
        _logger.critical(msg)
        raise click.BadParameter(f"Error: {msg}")

    if not is_valid_rule(rule):
        msg = f"Error parsing rule from user input: {single_rule}"
        _logger.critical(msg)
        raise click.BadParameter(f"Error: {msg}")

    _logger.debug("Processing rule: %s", rule["sid"])

    rule_report = analyze_rule(rule, checkers=checkers)

    write_output(OutputReport(rules=[rule_report]), out)


def process_rules_file(  # noqa: C901, PLR0912, PLR0915
    rules: str,
    evaluate_disabled: bool,
    checkers: Optional[Sequence[CheckerInterface]] = None,
) -> OutputReport:
    """Processes a rule file and returns a list of rules and their issues.

    Args:
    rules: A path to a Suricata rules file.
    evaluate_disabled: A flag indicating whether disabled rules should be evaluated.
    checkers: The checkers to be used when processing the rule file.

    Returns:
        A list of rules and their issues.

    Raises:
        RuntimeError: If no checkers could be automatically discovered.

    """
    if checkers is None:
        checkers = get_checkers()

    output = OutputReport()

    with (
        open(
            os.path.normpath(rules),
            buffering=io.DEFAULT_BUFFER_SIZE,
        ) as rules_fh,
    ):
        if len(checkers) == 0:
            msg = "No checkers provided for processing rules."
            _logger.error(msg)
            raise RuntimeError(msg)

        _logger.info("Processing rule file: %s", rules)

        collected_multiline_parts: Optional[str] = None
        multiline_begin_number: Optional[int] = None

        for number, line in enumerate(rules_fh.readlines(), start=1):
            # First work on collecting and parsing multiline rules
            if line.rstrip("\r\n").endswith("\\"):
                multiline_part = line.rstrip("\r\n")[:-1]

                if collected_multiline_parts is None:
                    collected_multiline_parts = multiline_part
                    multiline_begin_number = number
                else:
                    collected_multiline_parts += multiline_part.lstrip()

                continue

            # Process final part of multiline rule if one is being collected
            if collected_multiline_parts is not None:
                collected_multiline_parts += line.lstrip()

                rule_line = collected_multiline_parts.strip()

                collected_multiline_parts = None
            # If no multiline rule is being collected process as a potential single line rule
            else:
                if len(line.strip()) == 0:
                    continue

                if line.strip().startswith("#"):
                    if evaluate_disabled:
                        # Verify that this line is a rule and not a comment
                        if suricata_check.rule.parse(line) is None:
                            # Log the comment since it may be a invalid rule
                            _logger.warning(
                                "Ignoring comment on line %i: %s", number, line
                            )
                            continue
                    else:
                        # Skip the rule
                        continue

                rule_line = line.strip()

            try:
                rule: Optional[Rule] = suricata_check.rule.parse(rule_line)
            except suricata_check.rule.ParsingError:
                _logger.error(
                    "Internal error in parsing of rule on line %i: %s",
                    number,
                    rule_line,
                )
                rule = None

            # Parse comment and potential ignore comment to ignore rules
            ignore = __parse_type_ignore(rule)

            # Verify that a rule was parsed correctly.
            if rule is None:
                _logger.error("Error parsing rule on line %i: %s", number, rule_line)
                continue

            if not is_valid_rule(rule):
                _logger.error("Invalid rule on line %i: %s", number, rule_line)
                continue

            _logger.debug("Processing rule: %s on line %i", rule["sid"], number)

            rule_report: RuleReport = analyze_rule(
                rule,
                checkers=checkers,
                ignore=ignore,
            )
            rule_report.line_begin = multiline_begin_number or number
            rule_report.line_end = number

            output.rules.append(rule_report)

            multiline_begin_number = None

    _logger.info("Completed processing rule file: %s", rules)

    output.summary = summarize_output(output, checkers)

    return output


def __parse_type_ignore(rule: Optional[Rule]) -> Optional[Sequence[str]]:
    if rule is None:
        return None

    ignore_value = get_rule_suboption(rule, "metadata", "suricata-check")
    if ignore_value is None:
        return []

    return ignore_value.strip(' "').split(",")


def analyze_rule(
    rule: Rule,
    checkers: Optional[Sequence[CheckerInterface]] = None,
    ignore: Optional[Sequence[str]] = None,
) -> RuleReport:
    """Checks a rule and returns a dictionary containing the rule and a list of issues found.

    Args:
    rule: The rule to be checked.
    checkers: The checkers to be used to check the rule.
    ignore: Regular expressions to match checker codes to ignore

    Returns:
    A list of issues found in the rule.
    Each issue is typed as a `dict`.

    Raises:
    InvalidRuleError: If the rule does not follow the Suricata syntax.

    """
    if not is_valid_rule(rule):
        raise InvalidRuleError(rule["raw"])

    check_rule_option_recognition(rule)

    if checkers is None:
        checkers = get_checkers()

    rule_report: RuleReport = RuleReport(rule=rule)

    _logger.warning(ignore)

    compiled_ignore = (
        [_regex_provider.compile(r) for r in ignore] if ignore is not None else []
    )

    for checker in checkers:
        try:
            issues = checker.check_rule(rule)
            for r in compiled_ignore:
                issues = list(filter(lambda issue: r.match(issue.code) is None, issues))
            rule_report.add_issues(issues)
        except Exception as exception:  # noqa: BLE001
            _logger.warning(
                "Failed to run %s on rule: %s",
                checker.__class__.__name__,
                rule["raw"],
                extra={"exception": exception},
            )

    rule_report.summary = summarize_rule(rule_report, checkers)

    return rule_report


if __name__ == "__main__":
    main()



"""The `suricata_check.utils.regex` module contains regular expressions for matching various parts of rules."""

import importlib.util
import logging
from collections.abc import Iterable, Sequence
from functools import lru_cache

from suricata_check.utils.checker_typing import Rule

_logger = logging.getLogger(__name__)

# Import the fastest regex provider available:
if importlib.util.find_spec("regex") is not None:
    _logger.info("Detected regex module as installed, using it.")
    import regex as _regex_provider
else:
    _logger.warning(
        """Did not detect regex module as installed, using re instead.
To increase suricata-check processing speed, consider isntalling the regex module \
by running `pip install suricata-check[performance]`.""",
    )
    import re as _regex_provider

LRU_CACHE_SIZE = 10

ADDRESS_GROUPS = (
    "HOME_NET",
    "EXTERNAL_NET",
    "HTTP_SERVERS",
    "SMTP_SERVERS",
    "SQL_SERVERS",
    "DNS_SERVERS",
    "TELNET_SERVERS",
    "AIM_SERVERS",
    "DC_SERVERS",
    "DNP3_SERVER",
    "DNP3_CLIENT",
    "MODBUS_CLIENT",
    "MODBUS_SERVER",
    "ENIP_CLIENT",
    "ENIP_SERVER",
)


PORT_GROUPS = (
    "HTTP_PORTS",
    "SHELLCODE_PORTS",
    "ORACLE_PORTS",
    "SSH_PORTS",
    "DNP3_PORTS",
    "MODBUS_PORTS",
    "FILE_DATA_PORTS",
    "FTP_PORTS",
    "GENEVE_PORTS",
    "VXLAN_PORTS",
    "TEREDO_PORTS",
)

ALL_VARIABLES = ADDRESS_GROUPS + PORT_GROUPS

CLASSTYPES = (
    "not-suspicious",
    "unknown",
    "bad-unknown",
    "attempted-recon",
    "successful-recon-limited",
    "successful-recon-largescale",
    "attempted-dos",
    "successful-dos",
    "attempted-user",
    "unsuccessful-user",
    "successful-user",
    "attempted-admin",
    "successful-admin",
    # NEW CLASSIFICATIONS
    "rpc-portmap-decode",
    "shellcode-detect",
    "string-detect",
    "suspicious-filename-detect",
    "suspicious-login",
    "system-call-detect",
    "tcp-connection",
    "trojan-activity",
    "unusual-client-port-connection",
    "network-scan",
    "denial-of-service",
    "non-standard-protocol",
    "protocol-command-decode",
    "web-application-activity",
    "web-application-attack",
    "misc-activity",
    "misc-attack",
    "icmp-event",
    "inappropriate-content",
    "policy-violation",
    "default-login-attempt",
    # Update
    "targeted-activity",
    "exploit-kit",
    "external-ip-check",
    "domain-c2",
    "pup-activity",
    "credential-theft",
    "social-engineering",
    "coin-mining",
    "command-and-control",
)

NON_FUNCTIONAL_KEYWORDS = (
    "classtype",
    "gid",
    "metadata",
    "msg",
    "priority",
    "reference",
    "rev",
    "sid",
    "target",
)

FLOW_KEYWORDS = (
    "flow",
    "flow.age",
    "flowint",
)

STREAM_KEYWORDS = ("stream_size",)

FLOW_STREAM_KEYWORDS: Sequence[str] = tuple(
    sorted(set(FLOW_KEYWORDS).union(STREAM_KEYWORDS)),
)

STICKY_BUFFER_NAMING = {
    "dce_iface": "dce.iface",
    "dce_opnum": "dce.opnum",
    "dce_stub_data": "dce.stub_data",
    "dns_query": "dns.query",
    "file_data": "file.data",
    "http_accept": "http.accept",
    "http_accept_enc": "http.accept_enc",
    "http_accept_lang": "http.accept_lang",
    "http_client_body": "http.request_body",
    "http_connection": "http.connection",
    "http_content_len": "http.content_len",
    "http_content_type": "http.content_type",
    "http_cookie": "http.cookie",
    "http_header": "http.header",
    "http_header_names": "http.header_names",
    "http_host": "http.host",
    "http_method": "http.method",
    "http_protocol": "http.protocol",
    "http_raw_header": "http.header.raw",
    "http_raw_host": "http.host.raw",
    "http_raw_uri": "http.uri.raw",
    "http_referer": "http.referer",
    "http_request_line": "http.request_line",
    "http_response_line": "http.response_line",
    "http_server_body": "http.response_body",
    "http_start": "http.start",
    "http_stat_code": "http.stat_code",
    "http_stat_msg": "http.stat_msg",
    "http_uri": "http.uri",
    "http_user_agent": "http.user_agent",
    "ja3_hash": "ja3.hash",
    "tls_cert_fingerprint": "tls.cert_fingerprint",
    "tls_cert_issuer": "tls.cert_issuer",
    "tls_cert_serial": "tls.cert_serial",
    "tls_cert_subject": "tls.cert_subject",
    "tls_sni": "tls.sni",
}

BASE64_BUFFER_KEYWORDS = ("base64_data",)

OTHER_BUFFERS = (
    "http.location",
    "http.request_header",
    "http.response_header",
    "http.server",
    "ja3s.hash",
    "tls.certs",
    "tls.version",
)

assert set(OTHER_BUFFERS).isdisjoint(
    set(STICKY_BUFFER_NAMING.keys()).union(STICKY_BUFFER_NAMING.values())
)

BUFFER_KEYWORDS: Sequence[str] = tuple(
    sorted(
        set(STICKY_BUFFER_NAMING.keys())
        .union(STICKY_BUFFER_NAMING.values())
        .union(BASE64_BUFFER_KEYWORDS)
        .union(OTHER_BUFFERS),
    ),
)

SIZE_KEYWORDS = (
    "bsize",
    "dsize",
)

TRANSFORMATION_KEYWORDS = (
    "compress_whitespace",
    "dotprefix",
    "header_lowercase",
    "pcrexform",
    "strip_pseudo_headers",
    "strip_whitespace",
    "to_lowercase",
    "to_md5",
    "to_sha1",
    "to_sha256",
    "to_uppercase",
    "url_decode",
    "xor",
)

BASE64_TRANSFORMATION_KEYWORDS = ("base64_decode",)

ALL_TRANSFORMATION_KEYWORDS: Sequence[str] = tuple(
    sorted(set(TRANSFORMATION_KEYWORDS).union(BASE64_TRANSFORMATION_KEYWORDS)),
)

CONTENT_KEYWORDS = ("content", "pcre")

POINTER_MOVEMENT_KEYWORDS = (
    "depth",
    "distance",
    "offset",
    "pkt_data",
    "within",
)

COMPATIBILITY_MODIFIER_KEYWORDS = ("rawbytes",)

MODIFIER_KEYWORDS = ("nocase",)

ALL_MODIFIER_KEYWORDS: Sequence[str] = tuple(
    sorted(set(COMPATIBILITY_MODIFIER_KEYWORDS).union(MODIFIER_KEYWORDS)),
)

MATCH_LOCATION_KEYWORDS = (
    "endswith",
    "startswith",
)

OTHER_PAYLOAD_KEYWORDS = (
    "byte_extract",
    "byte_jump",
    "byte_test",
    "isdataat",
)

IP_SPECIFIC_KEYWORDS = (
    "ip_proto",
    "ttl",
)

TCP_SPECIFIC_KEYWORDS = (
    "ack",
    "flags",  # This is a duplicate of tcp.flags
    "seq",
    "tcp.flags",
    "tcp.hdr",
)

UDP_SPECIFIC_KEYWORDS = ("udp.hdr",)

ICMP_SPECIFIC_KEYWORDS = (
    "fragbits",
    "icode",
    "icmp_id",
    "icmp_seq",
    "itype",
)

HTTP_SPECIFIC_KEYWORDS = (
    "file.data",
    "file_data",
    "http.accept",
    "http.accept_enc",
    "http.accept_lang",
    "http.connection",
    "http.content_len",
    "http.content_len",
    "http.content_type",
    "http.cookie",
    "http.header",
    "http.header_names",
    "http.header.raw",
    "http.host",
    "http.host.raw",
    "http.location",
    "http.method",
    "http.protocol",
    "http.referer",
    "http.request_body",
    "http.request_header",
    "http.request_line",
    "http.response_body",
    "http.response_header",
    "http.response_line",
    "http.server",
    "http.start",
    "http.stat_code",
    "http.stat_code",
    "http.stat_msg",
    "http.uri",
    "http.uri.raw",
    "http.user_agent",
    "http_accept",
    "http_accept_enc",
    "http_accept_lang",
    "http_connection",
    "http_content_len",
    "http_content_len",
    "http_content_type",
    "http_cookie",
    "http_header",
    "http_header_names",
    "http_host",
    "http_location",
    "http_method",
    "http_protocol",
    "http_raw_header",
    "http_raw_host",
    "http_raw_uri",
    "http_referer",
    "http_request_line",
    "http_response_line",
    "http_server_body",
    "http_start",
    "http_stat_code",
    "http_stat_msg",
    "http_uri",
    "http_user_agent",
    "urilen",
)

DNS_SPECIFIC_KEYWORDS = (
    "dns.opcode",
    "dns.query",
    "dns_query",
)

TLS_SPECIFIC_KEYWORDS = (
    "ssl_version",
    "ssl_state",
    "tls.cert_fingerprint",
    "tls.cert_issuer",
    "tls.cert_serial",
    "tls.cert_subject",
    "tls.certs",
    "tls.sni",
    "tls.version",
    "tls_cert_fingerprint",
    "tls_cert_issuer",
    "tls_cert_serial",
    "tls_cert_subject",
    "tls_sni",
)

SSH_SPECIFIC_KEYWORDS = ("ssh_proto",)

JA3_JA4_KEYWORDS = (
    "ja3.hash",
    "ja3_hash",
    "ja3.string",
    "ja3s.hash",
)

DCERPC_SPECIFIC_KEYWORDS = (
    "dce.iface",
    "dce.opnum",
    "dce.stub_data",
    "dce_iface",
    "dce_opnum",
    "dce_stub_data",
)

FTP_KEYWORDS = ("ftpbounce", "ftpdata_command")

APP_LAYER_KEYWORDS = (
    "app-layer-event",
    "app-layer-protocol",
)

PROTOCOL_SPECIFIC_KEYWORDS = tuple(
    sorted(
        set().union(
            *(
                IP_SPECIFIC_KEYWORDS,
                TCP_SPECIFIC_KEYWORDS,
                UDP_SPECIFIC_KEYWORDS,
                ICMP_SPECIFIC_KEYWORDS,
                HTTP_SPECIFIC_KEYWORDS,
                DNS_SPECIFIC_KEYWORDS,
                TLS_SPECIFIC_KEYWORDS,
                SSH_SPECIFIC_KEYWORDS,
                DCERPC_SPECIFIC_KEYWORDS,
                JA3_JA4_KEYWORDS,
                FTP_KEYWORDS,
                APP_LAYER_KEYWORDS,
            ),
        ),
    ),
)

PERFORMANCE_DETECTION_OPTIONS = ("fast_pattern",)

LUA_KEYWORDS = ("lua", "luajit")

ALL_DETECTION_KEYWORDS: Sequence[str] = tuple(
    sorted(
        set().union(
            *(
                BUFFER_KEYWORDS,
                SIZE_KEYWORDS,
                ALL_TRANSFORMATION_KEYWORDS,
                CONTENT_KEYWORDS,
                POINTER_MOVEMENT_KEYWORDS,
                ALL_MODIFIER_KEYWORDS,
                MATCH_LOCATION_KEYWORDS,
                OTHER_PAYLOAD_KEYWORDS,
                PROTOCOL_SPECIFIC_KEYWORDS,
                PERFORMANCE_DETECTION_OPTIONS,
                LUA_KEYWORDS,
            ),
        ),
    ),
)

THRESHOLD_KEYWORDS = (
    "detection_filter",
    "threshold",
)

STATEFUL_KEYWORDS = ("flowbits", "flowint", "xbits")

OTHER_KEYWORDS = ("noalert", "tag")

ALL_KEYWORDS = tuple(
    sorted(
        set().union(
            *(
                NON_FUNCTIONAL_KEYWORDS,
                FLOW_KEYWORDS,
                STREAM_KEYWORDS,
                ALL_DETECTION_KEYWORDS,
                THRESHOLD_KEYWORDS,
                STATEFUL_KEYWORDS,
                OTHER_KEYWORDS,
            ),
        ),
    ),
)

METADATA_DATE_KEYWORDS = (
    "created_at",
    "reviewed_at",
    "updated_at",
)

METADATA_NON_DATE_KEYWORDS = (
    "affected_product",
    "attack_target",
    "confidence",
    "cve",
    "deprecation_reason",
    "deployment",
    "former_category",
    "former_sid",
    "impact_flag",
    "malware_family",
    "mitre_tactic_id",
    "mitre_tactic_name",
    "mitre_technique_id",
    "mitre_technique_name",
    "performance_impact",
    "policy",
    "ruleset",
    "signature_severity",
    "tag",
    "tls_state",
    "first_seen",
    "confidence_level",
)

ALL_METADATA_KEYWORDS = tuple(
    sorted(set(METADATA_DATE_KEYWORDS).union(METADATA_NON_DATE_KEYWORDS)),
)

IP_ADDRESS_REGEX = _regex_provider.compile(r"^.*\d+\.\d+\.\d+\.\d+.*$")

_GROUP_REGEX = _regex_provider.compile(r"^(!)?\[(.*)\]$")
_VARIABLE_GROUP_REGEX = _regex_provider.compile(r"^!?\$([A-Z\_]+)$")

_ACTION_REGEX = _regex_provider.compile(
    r"(alert|pass|drop|reject|rejectsrc|rejectdst|rejectboth)",
)
_PROTOCOL_REGEX = _regex_provider.compile(r"[a-z0-3\-]+")
_ADDR_REGEX = _regex_provider.compile(r"[a-zA-Z0-9\$_\!\[\],\s/\.]+")
_PORT_REGEX = _regex_provider.compile(r"[a-zA-Z0-9\$_\!\[\],\s:]+")
_DIRECTION_REGEX = _regex_provider.compile(r"(\->|<>)")
HEADER_REGEX = _regex_provider.compile(
    rf"{_ACTION_REGEX.pattern}\s*{_PROTOCOL_REGEX.pattern}\s*{_ADDR_REGEX.pattern}\s*{_PORT_REGEX.pattern}\s*{_DIRECTION_REGEX.pattern}\s*{_ADDR_REGEX.pattern}\s*{_PORT_REGEX.pattern}",
)
_OPTION_REGEX = _regex_provider.compile(
    r"[a-z\-\._]+\s*(:(\s*([0-9]+|.+)\s*\,?\s*)+)?;"
)
_BODY_REGEX = _regex_provider.compile(rf"\((\s*{_OPTION_REGEX.pattern}\s*)*\)")
_RULE_REGEX = _regex_provider.compile(
    rf"^(\s*#)?\s*{HEADER_REGEX.pattern}\s*{_BODY_REGEX.pattern}\s*(#.*)?$",
)


def get_regex_provider():  # noqa: ANN201
    """Returns the regex provider to be used.

    If `regex` is installed, it will return that module.
    Otherwise, it will return the `re` module instead.
    """
    return _regex_provider


@lru_cache(maxsize=LRU_CACHE_SIZE)
def __escape_regex(s: str) -> str:
    # Escape the escape character first
    s = s.replace("\\", "\\\\")

    # Then escape all other characters
    # . ^ $ * + ? { } [ ] \ | ( )
    s = s.replace(".", "\\.")
    s = s.replace("^", "\\^")
    s = s.replace("$", "\\$")
    s = s.replace("*", "\\*")
    s = s.replace("+", "\\+")
    s = s.replace("?", "\\?")
    s = s.replace("{", "\\{")
    s = s.replace("}", "\\}")
    s = s.replace("[", "\\[")
    s = s.replace("]", "\\]")
    s = s.replace("|", "\\|")
    s = s.replace("(", "\\(")
    s = s.replace(")", "\\)")

    return s  # noqa: RET504


def get_options_regex(options: Iterable[str]) -> _regex_provider.Pattern:
    """Returns a regular expression that can match any of the provided options."""
    return __get_options_regex(tuple(sorted(options)))


@lru_cache(maxsize=LRU_CACHE_SIZE)
def __get_options_regex(options: Sequence[str]) -> _regex_provider.Pattern:
    return _regex_provider.compile(
        "(" + "|".join([__escape_regex(option) for option in options]) + ")",
    )


def __is_group(entry: str) -> bool:
    if _GROUP_REGEX.match(entry) is None:
        return False

    return True


def get_rule_group_entries(group: str) -> Sequence[str]:
    """Returns a list of entries in a group."""
    stripped_group = group.strip()

    if not __is_group(stripped_group):
        return [stripped_group]

    match = _GROUP_REGEX.match(stripped_group)
    assert match is not None
    negated = match.group(1) == "!"

    entries = []
    for entry in match.group(2).split(","):
        stripped_entry = entry.strip()
        if __is_group(stripped_entry):
            entries += get_rule_group_entries(stripped_entry)
        else:
            entries.append(stripped_entry)

    if negated:
        entries = ["!" + entry for entry in entries]

    return entries


def get_variable_groups(value: str) -> Sequence[str]:
    """Returns a list of variable groups such as $HTTP_SERVERS in a variable."""
    return __get_variable_groups(value)


@lru_cache(maxsize=LRU_CACHE_SIZE)
def __get_variable_groups(value: str) -> Sequence[str]:
    entries = get_rule_group_entries(value)
    variable_groups = []
    for entry in entries:
        match = _VARIABLE_GROUP_REGEX.match(entry)
        if match is not None:
            variable_groups.append(match.group(1))

    return variable_groups


def get_rule_body(rule: Rule) -> str:
    """Returns the body of a rule."""
    return __get_rule_body(rule)


@lru_cache(maxsize=LRU_CACHE_SIZE)
def __get_rule_body(rule: Rule) -> str:
    match = _BODY_REGEX.search(rule["raw"])

    if match is None:
        msg = f"Could not extract rule body from rule: {rule['raw']}"
        _logger.critical(msg)
        raise RuntimeError(msg)

    return match.group(0)


def is_valid_rule(rule: Rule) -> bool:
    """Checks if a rule is valid."""
    if _RULE_REGEX.match(rule["raw"]) is None:
        return False

    return True



"""DummyChecker."""

import logging
from collections.abc import Iterable
from typing import Optional

from suricata_check.checkers.interface.checker import CheckerInterface

_logger = logging.getLogger(__name__)


class DummyChecker(CheckerInterface):
    """Dummy class to prevent runtime errors on import."""

    codes = {}
    enabled_by_default = False

    def __init__(self: "DummyChecker", include: Optional[Iterable[str]] = None) -> None:
        """Log an error due to failed imports for the checker."""
        _logger.warning(
            "Failed to initialize %s due to failed imports. Ensure all necessary dependencies are installed.",
            self.__class__.__name__,
        )
        super().__init__(include=include)



"""`GenericChecker`."""

import logging
import warnings
from functools import lru_cache
from typing import Optional

import pytest

from suricata_check.checkers.interface.checker import CheckerInterface
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue, Rule
from suricata_check.utils.regex import get_regex_provider

_regex_provider = get_regex_provider()

_CODE_STRUCTURE_REGEX = _regex_provider.compile(r"[A-Z]{1,}[0-9]{3}")


class GenericChecker:
    """The GenericChecker class can be extended by tests to test classes implementing `CheckerInterface`."""

    checker: CheckerInterface

    @pytest.fixture(autouse=True)
    def __run_around_tests(self: "GenericChecker") -> None:
        logging.basicConfig(level=logging.DEBUG)

    def _set_log_level(self: "GenericChecker", level: int) -> None:
        logger = logging.getLogger()
        logger.setLevel(level)
        for handler in logger.handlers:
            handler.setLevel(level)

    @lru_cache(maxsize=1)
    def _check_rule(
        self: "GenericChecker",
        rule: Rule,
    ) -> ISSUES_TYPE:
        return self.checker.check_rule(rule)

    def _test_issue(
        self: "GenericChecker",
        rule: Optional[Rule],
        code: str,
        raised: bool,
        fail: bool = True,
    ) -> Optional[bool]:
        """Checks whether a rule raises or does not raise an issue with a given code.

        Raises a pytest failure, warning, or returns a boolean based on the provided arguments.
        """
        if rule is None:
            pytest.fail("Rule is None")

        correct, issue = self.check_issue(rule, code, raised)

        if correct is not True:
            msg = f"""\
{'Unexpected' if not raised else 'Missing'} code {code}.
{rule['raw']}
{issue}\
"""
            if fail:
                pytest.fail(msg)
            else:
                warnings.warn(RuntimeWarning(msg))

    def check_issue(
        self: "GenericChecker",
        rule: Optional[Rule],
        code: str,
        raised: bool,
    ) -> tuple[Optional[bool], Optional[Issue]]:
        """Checks whether a rule raises an issue with a certain code and returns whether the expectation is met."""
        if rule is None:
            pytest.fail("Rule is None")

        issues: ISSUES_TYPE = self._check_rule(rule)

        self._test_no_undeclared_codes(issues)
        self._test_issue_metadata(issues)

        correct: Optional[bool] = None
        issue: Optional[Issue] = None

        if raised:
            correct = False
            for issue in issues:
                if issue.code == code:
                    correct = True
                    break
            issue = None
        elif not raised:
            correct = True
            for issue in issues:
                if issue.code == code:
                    correct = False
                    break

        return correct, issue if not correct else None

    def _test_no_undeclared_codes(self: "GenericChecker", issues: ISSUES_TYPE) -> None:
        """Asserts the checker emits no undeclared codes."""
        assert self.checker is not None

        codes = set()
        for issue in issues:
            codes.add(issue.code)

        for code in codes:
            if code not in self.checker.codes:
                pytest.fail(code)

    @pytest.hookimpl(trylast=True)
    def test_code_structure(self: "GenericChecker") -> None:
        """Asserts the checker only emits codes following the allowed structure."""
        for code in self.checker.codes:
            if _CODE_STRUCTURE_REGEX.match(code) is None:
                pytest.fail(code)

    def _test_issue_metadata(self: "GenericChecker", issues: ISSUES_TYPE) -> None:
        """Asserts the checker adds required metadata to emitted issues."""
        for issue in issues:
            if not hasattr(issue, "checker"):
                pytest.fail(
                    "Issue with code {} did not specify checker.".format(
                        str(issue.code)
                    )
                )
            if not hasattr(issue, "severity"):
                pytest.fail(
                    "Issue with code {} did not specify severity.".format(
                        str(issue.code)
                    )
                )
            if issue.message.strip() != issue.message:
                pytest.fail(
                    'Issue with code {} starts with or ends with whitespace in message: """{}"""'.format(
                        str(issue.code), str(issue.message)
                    )
                )



"""The `suricata_check.checkers.styleguide` modules contains several checkers based on the Suricata Style Guide.

Reference: https://github.com/sidallocation/suricata-style-guide
"""

from suricata_check.checkers.styleguide.metadata import MetadataChecker
from suricata_check.checkers.styleguide.msg import MsgChecker
from suricata_check.checkers.styleguide.order import OrderChecker
from suricata_check.checkers.styleguide.overall import OverallChecker
from suricata_check.checkers.styleguide.pcre import PcreChecker
from suricata_check.checkers.styleguide.performance import PerformanceChecker
from suricata_check.checkers.styleguide.reference import ReferenceChecker
from suricata_check.checkers.styleguide.sid import SidChecker
from suricata_check.checkers.styleguide.state import StateChecker
from suricata_check.checkers.styleguide.whitespace import WhitespaceChecker

__all__ = [
    "MetadataChecker",
    "MsgChecker",
    "OrderChecker",
    "OverallChecker",
    "PcreChecker",
    "PerformanceChecker",
    "ReferenceChecker",
    "SidChecker",
    "StateChecker",
    "WhitespaceChecker",
]



"""The `suricata_check.checkers` module contains all rule checkers."""

from suricata_check.checkers import interface
from suricata_check.checkers.community import BestChecker, UnexpectedChecker
from suricata_check.checkers.mandatory import MandatoryChecker
from suricata_check.checkers.styleguide import (
    MetadataChecker,
    MsgChecker,
    OrderChecker,
    OverallChecker,
    PcreChecker,
    PerformanceChecker,
    ReferenceChecker,
    SidChecker,
    StateChecker,
    WhitespaceChecker,
)

__all__ = [
    "BestChecker",
    "MandatoryChecker",
    "MetadataChecker",
    "MsgChecker",
    "OrderChecker",
    "OverallChecker",
    "PcreChecker",
    "PerformanceChecker",
    "ReferenceChecker",
    "SidChecker",
    "StateChecker",
    "UnexpectedChecker",
    "WhitespaceChecker",
    "interface",
]



"""`OrderChecker`."""

import logging

from suricata_check.checkers.interface import CheckerInterface
from suricata_check.utils.checker import (
    are_rule_options_put_before,
    count_rule_options,
    get_rule_keyword_sequences,
    get_rule_option_position,
    is_rule_option_always_put_before,
    is_rule_option_first,
    is_rule_option_put_before,
    is_rule_option_set,
)
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue, Rule
from suricata_check.utils.regex import (
    ALL_DETECTION_KEYWORDS,
    ALL_TRANSFORMATION_KEYWORDS,
    BUFFER_KEYWORDS,
    CONTENT_KEYWORDS,
    FLOW_STREAM_KEYWORDS,
    MODIFIER_KEYWORDS,
    OTHER_PAYLOAD_KEYWORDS,
    PERFORMANCE_DETECTION_OPTIONS,
    POINTER_MOVEMENT_KEYWORDS,
    SIZE_KEYWORDS,
    TRANSFORMATION_KEYWORDS,
    get_regex_provider,
    get_rule_body,
)

_regex_provider = get_regex_provider()


# Regular expressions are placed here such that they are compiled only once.
# This has a significant impact on the performance.
REGEX_S210 = _regex_provider.compile(
    r"^\(.*content\s*:.*;\s*content\s*:.*;.*(depth|offset)\s*:.*\)$",
)


class OrderChecker(CheckerInterface):
    """The `OrderChecker` contains several checks on the ordering Suricata options.

    Note that the correct ordering of detection options is as follows:
        1. Buffer
        2. Size
        3. Transformation
        4. Content
        5. Pointer movement
        6. Fast pattern
        7. Nocase
        8. Other payload options

    Codes S200-S209 report on the non-standard ordering of common options.

    Codes S210-S219 report on the non-standard ordering of content matches.

    Codes S220-S229 report on the non-standard ordering of flow options.

    Codes S230-S239 report on the non-standard ordering of detection options.

    Codes S240-S249 report on the non-standard ordering of threshold options.
    """

    codes = {
        "S200": {"severity": logging.INFO},
        "S201": {"severity": logging.INFO},
        "S202": {"severity": logging.INFO},
        "S203": {"severity": logging.INFO},
        "S204": {"severity": logging.INFO},
        "S205": {"severity": logging.INFO},
        "S206": {"severity": logging.INFO},
        "S207": {"severity": logging.INFO},
        "S208": {"severity": logging.INFO},
        "S210": {"severity": logging.INFO},
        "S211": {"severity": logging.INFO},
        "S212": {"severity": logging.INFO},
        "S220": {"severity": logging.INFO},
        "S221": {"severity": logging.INFO},
        "S222": {"severity": logging.INFO},
        "S223": {"severity": logging.INFO},
        "S224": {"severity": logging.INFO},
        "S230": {"severity": logging.INFO},
        "S231": {"severity": logging.INFO},
        "S232": {"severity": logging.INFO},
        "S233": {"severity": logging.INFO},
        "S234": {"severity": logging.INFO},
        "S235": {"severity": logging.INFO},
        "S236": {"severity": logging.INFO},
        "S240": {"severity": logging.INFO},
        "S241": {"severity": logging.INFO},
    }

    def _check_rule(  # noqa: C901, PLR0912, PLR0915
        self: "OrderChecker",
        rule: Rule,
    ) -> ISSUES_TYPE:
        issues: ISSUES_TYPE = []

        body = get_rule_body(rule)

        if is_rule_option_first(rule, "msg") is not True:
            issues.append(
                Issue(
                    code="S200",
                    message="""The rule body does not have msg as the first option.
Consider reording to make msg the first option.""",
                )
            )

        if is_rule_option_put_before(rule, "reference", ("content", "pcre")) is True:
            issues.append(
                Issue(
                    code="S201",
                    message="""The rule body contains the reference option before the detection logic.
Consider reording to put the detection logic directly after the msg option.""",
                )
            )

        if is_rule_option_put_before(rule, "classtype", ("reference",)) is True:
            issues.append(
                Issue(
                    code="S202",
                    message="""The rule body contains the classtype option before the reference option.
Consider reording to put the classtype option directly after the reference option.""",
                )
            )

        if is_rule_option_put_before(rule, "classtype", ("content", "pcre")) is True:
            issues.append(
                Issue(
                    code="S203",
                    message="""The rule body contains the classtype option before the detection logic.
Consider reording to put the classtype option directly after the detection logic.""",
                )
            )

        if is_rule_option_put_before(rule, "sid", ("classtype",)) is True:
            issues.append(
                Issue(
                    code="S204",
                    message="""The rule body contains the sid option before the classtype option.
Consider reording to put the sid option directly after the classtype option.""",
                )
            )

        if is_rule_option_put_before(rule, "sid", ("reference",)) is True:
            issues.append(
                Issue(
                    code="S205",
                    message="""The rule body contains the sid option before the reference option.
Consider reording to put the sid option directly after the reference option.""",
                )
            )

        if is_rule_option_put_before(rule, "sid", ("content", "pcre")) is True:
            issues.append(
                Issue(
                    code="S206",
                    message="""The rule body contains the sid option before the detection logic.
Consider reording to put the sid option directly after the detection logic.""",
                )
            )

        if is_rule_option_put_before(rule, "rev", ("sid",)) is True:
            issues.append(
                Issue(
                    code="S207",
                    message="""The rule body contains the rev option before the sid option.
Consider reording to put the rev option directly after the sid option.""",
                )
            )

        if is_rule_option_put_before(rule, "metadata", ("sid", "rev")) is True:
            issues.append(
                Issue(
                    code="S208",
                    message="""The rule body contains does not have the metadata option as the last option.
Consider making metadata the last option.""",
                )
            )

        if (
            REGEX_S210.match(
                body,
            )
            is not None
        ):
            issues.append(
                Issue(
                    code="S210",
                    message="""The rule body contains a content matches modified by depth or offset \
that is not the first content match.
Consider moving the modified content match to the beginning of the detection options.""",
                )
            )

        if count_rule_options(rule, "depth") > 1:
            issues.append(
                Issue(
                    code="S211",
                    message="""The rule body contains more than one content matche modified by depth.
Consider making the second content match relative to the first using the within option.""",
                )
            )

        if count_rule_options(rule, "offset") > 1:
            issues.append(
                Issue(
                    code="S212",
                    message="""The rule body contains more than one content matche modified by offset.
Consider making the second content match relative to the first using the distance option.""",
                )
            )

        if (
            is_rule_option_set(rule, "flow")
            and get_rule_option_position(rule, "flow") != 1
        ):
            issues.append(
                Issue(
                    code="S220",
                    message="""The rule flow option is set but not directly following the msg option.
Consider moving the flow option to directly after the msg option.""",
                )
            )

        if (
            is_rule_option_always_put_before(
                rule,
                "flow",
                FLOW_STREAM_KEYWORDS,
            )
            is False
        ):
            issues.append(
                Issue(
                    code="S221",
                    message="""The rule contains flow or stream keywords before the flow option in the rule body.
Consider moving the flow option to before the flow and/or stream keywords.""",
                )
            )

        if (
            are_rule_options_put_before(
                rule,
                ("content", "pcre"),
                FLOW_STREAM_KEYWORDS,
            )
            is True
        ):
            issues.append(
                Issue(
                    code="S222",
                    message="""The rule contains flow or stream keywords after content buffers or detection logic.
Consider moving the flow and/or stream keywords to before content buffers and detection options.""",
                )
            )

        if (
            is_rule_option_put_before(
                rule,
                "urilen",
                FLOW_STREAM_KEYWORDS,
            )
            is True
        ):
            issues.append(
                Issue(
                    code="S223",
                    message="""The rule contains the urilen option before the flow or stream keywords in the rule body.
Consider moving the urilen option to after the flow and/or stream keywords.""",
                )
            )

        if (
            is_rule_option_always_put_before(
                rule,
                "urilen",
                ("content", "pcre"),
            )
            is False
        ):
            issues.append(
                Issue(
                    code="S224",
                    message="""The rule contains the urilen option after content buffers or detection logic.
Consider moving the urilen option to before content buffers and detection options.""",
                )
            )

        # Detects pointer movement before any content or buffer option or between a buffer and a content option.
        for sequence in get_rule_keyword_sequences(
            rule, seperator_keywords=CONTENT_KEYWORDS
        ):
            if (
                are_rule_options_put_before(
                    rule,
                    POINTER_MOVEMENT_KEYWORDS,
                    set(CONTENT_KEYWORDS).union(BUFFER_KEYWORDS),
                    sequence=sequence,
                )
                is True
            ):
                issues.append(
                    Issue(
                        code="S230",
                        message="""The rule contains pointer movement before the content option in sequence {}.
Consider moving the pointer movement options to after the content option.""".format(
                            sequence
                        ),
                    )
                )

        # Detects fast_pattern before any content or buffer option or between a buffer and a content option.
        for sequence in get_rule_keyword_sequences(
            rule, seperator_keywords=CONTENT_KEYWORDS
        ):
            if (
                is_rule_option_put_before(
                    rule,
                    "fast_pattern",
                    set(SIZE_KEYWORDS)
                    .union(ALL_TRANSFORMATION_KEYWORDS)
                    .union(CONTENT_KEYWORDS)
                    .union(POINTER_MOVEMENT_KEYWORDS),
                    sequence=sequence,
                )
                is True
            ):
                issues.append(
                    Issue(
                        code="S231",
                        message="""The rule contains the fast_pattern option before \
size options, transformation options, the content option or pointer movement options in sequence {}.
Consider moving the fast_pattern option to after \
size options, transformation options, the content option or pointer movement options.""".format(
                            sequence
                        ),
                    )
                )

        # Detects no_case before any content or buffer option or between a buffer and a content option.
        for sequence in get_rule_keyword_sequences(
            rule, seperator_keywords=CONTENT_KEYWORDS
        ):
            if (
                is_rule_option_put_before(
                    rule,
                    "nocase",
                    set(SIZE_KEYWORDS)
                    .union(ALL_TRANSFORMATION_KEYWORDS)
                    .union(CONTENT_KEYWORDS)
                    .union(POINTER_MOVEMENT_KEYWORDS)
                    .union(PERFORMANCE_DETECTION_OPTIONS),
                    sequence=sequence,
                )
                is True
            ):
                issues.append(
                    Issue(
                        code="S232",
                        message="""The rule contains the nocase option before \
size options, transformation options, the content option, pointer movement options, or fast_pattern option in sequence {}.
Consider moving the nocase option to after \
size options, transformation options, the content option, pointer movement options, or fast_pattern option.""".format(
                            sequence
                        ),
                    )
                )

        # Detects modifier options before any content or buffer option or between a buffer and a content option.
        for sequence in get_rule_keyword_sequences(
            rule, seperator_keywords=CONTENT_KEYWORDS
        ):
            if (
                are_rule_options_put_before(
                    rule,
                    MODIFIER_KEYWORDS,
                    set(CONTENT_KEYWORDS),
                    sequence=sequence,
                )
                is True
            ):
                issues.append(
                    Issue(
                        code="S233",
                        message="""The rule contains modifier options before the content option.
Consider moving the modifier options to after the content option.""",
                    )
                )

        # Detects other detection options before any content or buffer option or between a buffer and a content option.
        for sequence in get_rule_keyword_sequences(
            rule, seperator_keywords=CONTENT_KEYWORDS
        ):
            if (
                are_rule_options_put_before(
                    rule,
                    OTHER_PAYLOAD_KEYWORDS,
                    set(CONTENT_KEYWORDS).union(BUFFER_KEYWORDS),
                    sequence=sequence,
                )
                is True
            ):
                issues.append(
                    Issue(
                        code="S234",
                        message="""The rule contains other detection options before \
size options, transformation options, the content option, pointer movement options, nocase option, or fast_pattern option.
Consider moving the other detection options to after \
size options, transformation options,  the content option, pointer movement options, nocase option, or fast_pattern option.""",
                    )
                )

        # Detects size options after any transformation options, content option or other detection options.
        for sequence in get_rule_keyword_sequences(
            rule, seperator_keywords=CONTENT_KEYWORDS
        ):
            if (
                are_rule_options_put_before(
                    rule,
                    set(TRANSFORMATION_KEYWORDS)
                    .union(CONTENT_KEYWORDS)
                    .union(OTHER_PAYLOAD_KEYWORDS),
                    SIZE_KEYWORDS,
                    sequence=sequence,
                )
                is True
            ):
                issues.append(
                    Issue(
                        code="S235",
                        message="""The rule contains other size options after \
any transformation options, content option or other detection options.
Consider moving the size options to after any transformation options, content option or other detection options""",
                    )
                )

        # Detects transformation options after any content option or other detection options.
        for sequence in get_rule_keyword_sequences(
            rule, seperator_keywords=CONTENT_KEYWORDS
        ):
            if (
                are_rule_options_put_before(
                    rule,
                    set(CONTENT_KEYWORDS).union(OTHER_PAYLOAD_KEYWORDS),
                    TRANSFORMATION_KEYWORDS,
                    sequence=sequence,
                )
                is True
            ):
                issues.append(
                    Issue(
                        code="S236",
                        message="""The rule contains other transformation options after \
any content option or other detection options.
Consider moving the transformation options to after any content option or other detection options""",
                    )
                )

        if (
            is_rule_option_put_before(
                rule,
                "threshold",
                ALL_DETECTION_KEYWORDS,
            )
            is True
        ):
            issues.append(
                Issue(
                    code="S240",
                    message="""The rule contains the threshold option before some detection option.
Consider moving the threshold option to after the detection options.""",
                )
            )

        if (
            is_rule_option_always_put_before(
                rule,
                "threshold",
                ("reference", "sid"),
            )
            is False
        ):
            issues.append(
                Issue(
                    code="S241",
                    message="""The rule contains the threshold option after the reference and/or sid option.
Consider moving the threshold option to before the reference and sid options.""",
                )
            )

        return issues



"""The `suricata_check.typing` module contains all types used by the `suricata-check` package."""

import json
from collections.abc import Iterable, MutableMapping, MutableSequence
from dataclasses import dataclass, field
from typing import (
    Optional,
    TypeVar,
)

from suricata_check.rule import Rule


class InvalidRuleError(RuntimeError):
    """Raised when an invalid rule is detected.

    Note that some rules may be invalid due to not following the Suricata rule syntax.
    Rules following the syntax, but considered invalid by Suricata due to missing options need not raise this error.
    Rules for which this error is not raised are not neccessarily syntactically correct but can be processed by suricata-check.
    """

    def __init__(self: "InvalidRuleError", message: str) -> None:
        """Initializes the `InvalidRuleError` with the raw rule as message."""
        super().__init__(message)


@dataclass
class Issue:
    """The `Issue` dataclass represents a single issue found in a rule."""

    code: str
    message: str
    severity: Optional[int] = None
    checker: Optional[str] = None

    def to_dict(self: "Issue") -> dict[str, str]:
        """Returns the Issue represented as a dictionary."""
        d = {
            "code": self.code,
            "message": self.message,
        }

        if self.checker is not None:
            d["checker"] = self.checker

        return d

    @property
    def hash(self: "Issue") -> int:
        """Returns a unique hash that can be used as a fingerprint for the issue."""
        return hash(tuple(sorted(self.to_dict().items())))

    def __repr__(self: "Issue") -> str:
        """Returns the Issue represented as a string."""
        return json.dumps(self.to_dict())


ISSUES_TYPE = MutableSequence[Issue]
SIMPLE_SUMMARY_TYPE = MutableMapping[str, int]
RULE_SUMMARY_TYPE = SIMPLE_SUMMARY_TYPE
EXTENSIVE_SUMMARY_TYPE = MutableMapping[str, SIMPLE_SUMMARY_TYPE]

Cls = TypeVar("Cls")


def get_all_subclasses(cls: type[Cls]) -> Iterable[type[Cls]]:
    """Returns all class types that subclass the provided type."""
    all_subclasses = []

    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses


@dataclass
class RuleReport:
    """The `RuleReport` dataclass represents a rule, together with information on its location and detected issues."""

    rule: Rule
    summary: Optional[RULE_SUMMARY_TYPE] = None
    line_begin: Optional[int] = None
    line_end: Optional[int] = None

    _issues: ISSUES_TYPE = field(default_factory=list, init=False)

    @property
    def issues(self: "RuleReport") -> ISSUES_TYPE:
        """List of issues found in the rule."""
        return self._issues

    def add_issue(self: "RuleReport", issue: Issue) -> None:
        """Adds an issue to the report."""
        self._issues.append(issue)

    def add_issues(self: "RuleReport", issues: ISSUES_TYPE) -> None:
        """Adds an issue to the report."""
        for issue in issues:
            self._issues.append(issue)

    def to_dict(self: "RuleReport") -> dict[str, str]:
        """Returns the RuleReport represented as a dictionary."""
        d = {
            "rule": self.rule["raw"],
            "issues": [issue.to_dict() for issue in self.issues],
        }

        if self.summary is not None:
            d["summary"] = self.summary

        if self.line_begin is not None or self.line_end is not None:
            d["lines"] = {}

        if self.line_begin is not None:
            d["lines"]["begin"] = self.line_begin

        if self.line_begin is not None:
            d["lines"]["end"] = self.line_end

        return d

    def __repr__(self: "RuleReport") -> str:
        """Returns the RuleReport represented as a string."""
        return json.dumps(self.to_dict())


RULE_REPORTS_TYPE = MutableSequence[RuleReport]


@dataclass
class OutputSummary:
    """The `OutputSummary` dataclass represent a collection of summaries on the output of `suricata_check`."""

    overall_summary: SIMPLE_SUMMARY_TYPE
    issues_by_group: SIMPLE_SUMMARY_TYPE
    issues_by_type: EXTENSIVE_SUMMARY_TYPE


@dataclass
class OutputReport:
    """The `OutputSummary` dataclass represent the `suricata_check`, consisting of rule reports and summaries."""

    _rules: RULE_REPORTS_TYPE = field(default_factory=list, init=False)
    summary: Optional[OutputSummary] = None

    def __init__(
        self: "OutputReport",
        rules: RULE_REPORTS_TYPE = [],
        summary: Optional[OutputSummary] = None,
    ) -> None:
        """Initialized the `OutputReport`, optionally with a list of rules and/or a summary."""
        self._rules = []
        for rule in rules:
            self.add_rule(rule)
        self.summary = summary
        super().__init__()

    @property
    def rules(self: "OutputReport") -> RULE_REPORTS_TYPE:
        """List of rules contained in the report."""
        return self._rules

    def add_rule(self: "OutputReport", rule_report: RuleReport) -> None:
        """Adds an rule to the report."""
        self._rules.append(rule_report)



"""The `suricata_check.utils.checker` module contains several utilities for developing rule checkers."""

import logging
from collections.abc import Iterable, Sequence
from functools import lru_cache
from typing import Optional, Union

from suricata_check.utils.checker_typing import Rule
from suricata_check.utils.regex import (
    ALL_DETECTION_KEYWORDS,
    ALL_KEYWORDS,
    ALL_METADATA_KEYWORDS,
    BUFFER_KEYWORDS,
    STICKY_BUFFER_NAMING,
    get_regex_provider,
    get_variable_groups,
)

_LRU_CACHE_SIZE = 10


_logger = logging.getLogger(__name__)

_regex_provider = get_regex_provider()


def check_rule_option_recognition(rule: Rule) -> None:
    """Checks whether all rule options and metadata options are recognized.

    Unrecognized options will be logged as a warning in `suricata-check.log`
    """
    for option in rule["options"]:
        name = option["name"]
        if name not in ALL_KEYWORDS:
            _logger.warning(
                "Option %s from rule %s is not recognized.",
                name,
                rule["sid"],
            )

    for option in rule["metadata"]:
        name = _regex_provider.split(r"\s+", option)[0]
        if name not in ALL_METADATA_KEYWORDS:
            _logger.warning(
                "Metadata option %s from rule %s is not recognized.",
                name,
                rule["sid"],
            )


def is_rule_option_set(rule: Rule, name: str) -> bool:
    """Checks whether a rule has a certain option set.

    Args:
        rule (suricata_check.rule.Rule): rule to be inspected
        name (str): name of the option

    Returns:
        bool: True iff the option is set atleast once

    """
    return __is_rule_option_set(rule, name)


@lru_cache(maxsize=_LRU_CACHE_SIZE)
def __is_rule_option_set(rule: Rule, name: str) -> bool:
    if name not in (
        "action",
        "proto",
        "source_addr",
        "source_port",
        "direction",
        "dest_addr",
        "dest_port",
    ):
        if name not in ALL_KEYWORDS:
            _logger.warning("Requested a non-recognized keyword: %s", name)

        for option in rule["options"]:
            if option["name"] == name:
                return True

        return False

    if name not in rule:
        return False

    if rule[name] is None:
        return False

    if rule[name] == "":
        return False

    return True


def get_rule_suboptions(
    rule: Rule, name: str, warn: bool = True
) -> Sequence[tuple[str, Optional[str]]]:
    """Returns a list of suboptions set in a rule."""
    values = get_rule_options(rule, name)
    valid_suboptions: list[tuple[str, Optional[str]]] = []
    for value in values:
        if value is None:
            continue
        values = value.split(",")
        suboptions: list[Optional[tuple[str, Optional[str]]]] = [
            __split_suboption(suboption, warn=warn) for suboption in values
        ]
        # Filter out suboptions that could not be parsed
        valid_suboptions += [
            suboption for suboption in suboptions if suboption is not None
        ]

    return valid_suboptions


def __split_suboption(
    suboption: str, warn: bool
) -> Optional[tuple[str, Optional[str]]]:
    suboption = suboption.strip()

    splitted = suboption.split(" ")
    splitted = [s.strip() for s in splitted]

    if len(splitted) == 1:
        return (splitted[0], None)
    if len(splitted) == 2:  # noqa: PLR2004
        return tuple(splitted)  # type: ignore reportReturnType

    if warn:
        _logger.warning("Failed to split suboption: %s", suboption)

    return None


def is_rule_suboption_set(rule: Rule, name: str, sub_name: str) -> bool:
    """Checks if a suboption within an option is set."""
    suboptions = get_rule_suboptions(rule, name)
    _logger.debug(suboptions)

    if sub_name in [suboption[0] for suboption in suboptions]:
        return True

    return False


def get_rule_suboption(rule: Rule, name: str, sub_name: str) -> Optional[str]:
    """Returns a suboption within an option is set."""
    suboptions = get_rule_suboptions(rule, name)

    for suboption in suboptions:
        if sub_name == suboption[0]:
            return suboption[1]

    msg = f"Option {name} not found in rule."
    _logger.debug(msg)
    return None


def count_rule_options(
    rule: Rule,
    name: Union[str, Iterable[str]],
) -> int:
    """Counts how often an option is set in a rule.

    Args:
        rule (suricata_check.rule.Rule): rule to be inspected
        name (Union[str, Iterable[str]]): name or names of the option

    Returns:
        int: The number of times an option is set

    """
    if not isinstance(name, str):
        name = tuple(sorted(name))
    return __count_rule_options(rule, name)


@lru_cache(maxsize=_LRU_CACHE_SIZE)
def __count_rule_options(
    rule: Rule,
    name: Union[str, Iterable[str]],
) -> int:
    count = 0

    if not isinstance(name, str):
        for single_name in name:
            count += count_rule_options(rule, single_name)
        return count

    if name not in (
        "action",
        "proto",
        "source_addr",
        "source_port",
        "direction",
        "dest_addr",
        "dest_port",
    ):
        if name not in ALL_KEYWORDS:
            _logger.warning("Requested a non-recognized keyword: %s", name)

        for option in rule["options"]:
            if option["name"] == name:
                count += 1

    if count == 0 and is_rule_option_set(rule, name):
        count = 1

    return count


def get_rule_option(rule: Rule, name: str) -> Optional[str]:
    """Retrieves one option of a rule with a certain name.

    If an option is set multiple times, it returns only one indeterminately.

    Args:
        rule (suricata_check.rule.Rule): rule to be inspected
        name (str): name of the option

    Returns:
        Optional[str]: The value of the option or None if it was not set.

    """
    return __get_rule_option(rule, name)


@lru_cache(maxsize=_LRU_CACHE_SIZE)
def __get_rule_option(rule: Rule, name: str) -> Optional[str]:
    options = get_rule_options(rule, name)

    if len(options) == 0:
        msg = f"Option {name} not found in rule."
        _logger.debug(msg)
        return None

    if len(options) == 0:
        msg = f"Cannot unambiguously determine the value of {name} because it is set multiple times."
        _logger.warning(msg)
        return None

    return options[0]


def get_rule_options(
    rule: Rule,
    name: Union[str, Iterable[str]],
) -> Sequence[Optional[str]]:
    """Retrieves all options of a rule with a certain name.

    Args:
        rule (suricata_check.rule.Rule): rule to be inspected
        name (Union[str, Iterable[str]]): name or names of the option

    Returns:
        Sequence[str]: The values of the option.

    """
    if not isinstance(name, str):
        name = tuple(sorted(name))
    return __get_rule_options(rule, name)


@lru_cache(maxsize=_LRU_CACHE_SIZE)
def __get_rule_options(
    rule: Rule,
    name: Union[str, Iterable[str]],
    warn_not_found: bool = True,
) -> Sequence[str]:
    values = []

    if not isinstance(name, str):
        for single_name in name:
            values.extend(__get_rule_options(rule, single_name, warn_not_found=False))
        return values

    if name not in (
        "action",
        "proto",
        "source_addr",
        "source_port",
        "direction",
        "dest_addr",
        "dest_port",
    ):
        if name not in ALL_KEYWORDS:
            _logger.warning("Requested a non-recognized keyword: %s", name)

        for option in rule["options"]:
            if option["name"] == name:
                values.append(option["value"])
    elif name in rule:
        values.append(rule[name])

    if warn_not_found and len(values) == 0:
        msg = f"Option {name} not found in rule {rule}."
        _logger.debug(msg)

    return values


def is_rule_option_equal_to(rule: Rule, name: str, value: str) -> bool:
    """Checks whether a rule has a certain option set to a certain value.

    If the option is set multiple times, it will return True if atleast one option matches the value.

    Args:
        rule (suricata_check.rule.Rule): rule to be inspected
        name (str): name of the option
        value (str): value to check for

    Returns:
        bool: True iff the rule has the option set to the value atleast once

    """
    if not is_rule_option_set(rule, name):
        return False

    values = get_rule_options(rule, name)

    for val in values:
        if val == value:
            return True

    return False


def is_rule_suboption_equal_to(
    rule: Rule, name: str, sub_name: str, value: str
) -> bool:
    """Checks whether a rule has a certain suboption set to a certain value.

    If the suboption is set multiple times, it will return True if atleast one option matches the value.

    Args:
        rule (suricata_check.rule.Rule): rule to be inspected
        name (str): name of the option
        sub_name (str): name of the suboption
        value (str): value to check for

    Returns:
        bool: True iff the rule has the option set to the value atleast once

    """
    if not is_rule_suboption_set(rule, name, sub_name):
        return False

    values = get_rule_suboptions(rule, name)

    for key, val in values:
        if key == sub_name and val == value:
            return True

    return False


def is_rule_option_equal_to_regex(
    rule: Rule,
    name: str,
    regex,  # re.Pattern or regex.Pattern  # noqa: ANN001
) -> bool:
    """Checks whether a rule has a certain option set to match a certain regex.

    If the option is set multiple times, it will return True if atleast one option matches the regex.

    Args:
        rule (suricata_check.rule.Rule): rule to be inspected
        name (str): name of the option
        regex (Union[re.Pattern, regex.Pattern]): regex to check for

    Returns:
        bool: True iff the rule has atleast one option matching the regex

    """
    if not is_rule_option_set(rule, name):
        return False

    values = get_rule_options(rule, name)

    for value in values:
        if value is None:
            continue
        if regex.match(value) is not None:
            return True

    return False


def is_rule_suboption_equal_to_regex(
    rule: Rule,
    name: str,
    sub_name: str,
    regex,  # re.Pattern or regex.Pattern  # noqa: ANN001
) -> bool:
    """Checks whether a rule has a certain option set to match a certain regex.

    If the option is set multiple times, it will return True if atleast one option matches the regex.

    Args:
        rule (suricata_check.rule.Rule): rule to be inspected
        name (str): name of the option
        sub_name (str): name of the suboption
        regex (Union[re.Pattern, regex.Pattern]): regex to check for

    Returns:
        bool: True iff the rule has atleast one option matching the regex

    """
    if not is_rule_suboption_set(rule, name, sub_name):
        return False

    values = get_rule_suboptions(rule, name)

    for key, value in values:
        if key == sub_name and regex.match(value) is not None:
            return True

    return False


def is_rule_option_always_equal_to_regex(
    rule: Rule,
    name: str,
    regex,  # re.Pattern or regex.Pattern  # noqa: ANN001
) -> Optional[bool]:
    """Checks whether a rule has a certain option set to match a certain regex.

    If the option is set multiple times, it will return True if all options match the regex.
    Returns none if the rule option is not set.

    Args:
        rule (suricata_check.rule.Rule): rule to be inspected
        name (str): name of the option
        regex (Union[re.Pattern, regex.Pattern]): regex to check for

    Returns:
        bool: True iff the rule has all options matching the regex

    """
    if not is_rule_option_set(rule, name):
        return None

    values = get_rule_options(rule, name)

    for value in values:
        if value is None:
            return False
        if regex.match(value) is None:
            return False

    return True


def is_rule_suboption_always_equal_to_regex(
    rule: Rule,
    name: str,
    sub_name: str,
    regex,  # re.Pattern or regex.Pattern  # noqa: ANN001
) -> Optional[bool]:
    """Checks whether a rule has a certain option set to match a certain regex.

    If the option is set multiple times, it will return True if all options match the regex.
    Returns none if the rule option is not set.

    Args:
        rule (suricata_check.rule.Rule): rule to be inspected
        name (str): name of the option
        sub_name (str): name of the suboption
        regex (Union[re.Pattern, regex.Pattern]): regex to check for

    Returns:
        bool: True iff the rule has all options matching the regex

    """
    if not is_rule_suboption_set(rule, name, sub_name):
        return None

    values = get_rule_suboptions(rule, name)

    for key, value in values:
        if key == sub_name and regex.match(value) is None:
            return False

    return True


def are_rule_options_equal_to_regex(
    rule: Rule,
    names: Iterable[str],
    regex,  # re.Pattern or regex.Pattern  # noqa: ANN001
) -> bool:
    """Checks whether a rule has certain options set to match a certain regex.

    If multiple options are set, it will return True if atleast one option matches the regex.

    Args:
        rule (suricata_check.rule.Rule): rule to be inspected
        names (Iterable[str]): names of the options
        regex (Union[re.Pattern, regex.Pattern]): regex to check for

    Returns:
        bool: True iff the rule has atleast one option matching the regex

    """
    for name in names:
        if is_rule_option_equal_to_regex(rule, name, regex):
            return True

    return False


def is_rule_option_one_of(
    rule: Rule,
    name: str,
    possible_values: Union[Sequence[str], set[str]],
) -> bool:
    """Checks whether a rule has a certain option set to a one of certain values.

    If the option is set multiple times, it will return True if atleast one option matches a value.

    Args:
        rule (suricata_check.rule.Rule): rule to be inspected
        name (str): name of the option
        possible_values (Iterable[str]): values to check for

    Returns:
        bool: True iff the rule has the option set to one of the values atleast once

    """
    if not is_rule_option_set(rule, name):
        return False

    values = get_rule_options(rule, name)

    for value in values:
        if value is None:
            continue
        if value in possible_values:
            return True

    return False


def get_rule_sticky_buffer_naming(
    rule: Rule,
) -> list[tuple[str, str]]:
    """Returns a list of tuples containing the name of a sticky buffer, and the modifier alternative."""
    sticky_buffer_naming = []
    for option in rule["options"]:
        if option["name"] in STICKY_BUFFER_NAMING:
            sticky_buffer_naming.append(
                (option["name"], STICKY_BUFFER_NAMING[option["name"]]),
            )

    return sticky_buffer_naming


def get_all_variable_groups(
    rule: Rule,
) -> list[str]:
    """Returns a list of variable groups such as $HTTP_SERVERS in a rule."""
    return __get_all_variable_groups(rule)


@lru_cache(maxsize=_LRU_CACHE_SIZE)
def __get_all_variable_groups(
    rule: Rule,
) -> list[str]:
    variable_groups = []
    for name in (
        "source_addr",
        "source_port",
        "direction",
        "dest_addr",
        "dest_port",
    ):
        if is_rule_option_set(rule, name):
            value = get_rule_option(rule, name)
            assert value is not None
            variable_groups += get_variable_groups(value)

    return variable_groups


def get_rule_option_positions(
    rule: Rule,
    name: str,
    sequence: Optional[tuple[str, ...]] = None,
) -> Sequence[int]:
    """Finds the positions of an option in the rule body.

    Optionally takes a sequence of options to use instead of `rule['options']`.
    """
    return __get_rule_option_positions(rule, name, sequence=sequence)


@lru_cache(maxsize=_LRU_CACHE_SIZE)
def __get_rule_option_positions(
    rule: Rule,
    name: str,
    sequence: Optional[tuple[str, ...]] = None,
) -> Sequence[int]:
    provided_sequence = True
    if sequence is None:
        sequence = tuple(option["name"] for option in rule["options"])
        provided_sequence = False

    positions = []
    for i, option in enumerate(sequence):
        if option == name:
            positions.append(i)

    if not provided_sequence and len(positions) == 0 and is_rule_option_set(rule, name):
        msg = f"Cannot determine position of {name} option since it is not part of the sequence of detection keywords."
        _logger.critical(msg)
        raise ValueError(msg)

    return tuple(sorted(positions))


def get_rule_option_position(rule: Rule, name: str) -> Optional[int]:
    """Finds the position of an option in the rule body.

    Return None if the option is not set or set multiple times.
    """
    return __get_rule_option_position(rule, name)


@lru_cache(maxsize=_LRU_CACHE_SIZE)
def __get_rule_option_position(rule: Rule, name: str) -> Optional[int]:
    positions = get_rule_option_positions(rule, name)

    if len(positions) == 0:
        _logger.debug(
            "Cannot unambigously determine the position of the %s option since it it not set.",
            name,
        )
        return None

    if len(positions) == 1:
        return positions[0]

    _logger.debug(
        "Cannot unambigously determine the position of the %s option since it is set multiple times.",
        name,
    )
    return None


def is_rule_option_first(rule: Rule, name: str) -> Optional[int]:
    """Checks if a rule option is positioned at the beginning of the body."""
    position = get_rule_option_position(rule, name)

    if position is None:
        _logger.debug("Cannot unambiguously determine if option %s first.", name)
        return None

    if position == 0:
        return True

    return False


def is_rule_option_last(rule: Rule, name: str) -> Optional[bool]:
    """Checks if a rule option is positioned at the end of the body."""
    position = get_rule_option_position(rule, name)

    if position is None:
        _logger.debug("Cannot unambiguously determine if option %s last.", name)
        return None

    if position == len(rule["options"]) - 1:
        return True

    return False


def get_rule_options_positions(
    rule: Rule,
    names: Iterable[str],
    sequence: Optional[Iterable[str]] = None,
) -> Iterable[int]:
    """Finds the positions of several options in the rule body."""
    return __get_rule_options_positions(
        rule, tuple(sorted(names)), sequence=tuple(sequence) if sequence else None
    )


@lru_cache(maxsize=_LRU_CACHE_SIZE)
def __get_rule_options_positions(
    rule: Rule,
    names: Iterable[str],
    sequence: Optional[tuple[str, ...]] = None,
) -> Iterable[int]:
    positions = []

    for name in names:
        positions.extend(get_rule_option_positions(rule, name, sequence=sequence))

    return tuple(sorted(positions))


def is_rule_option_put_before(
    rule: Rule,
    name: str,
    other_names: Union[Sequence[str], set[str]],
    sequence: Optional[Iterable[str]] = None,
) -> Optional[bool]:
    """Checks whether a rule option is placed before one or more other options."""
    return __is_rule_option_put_before(
        rule,
        name,
        tuple(sorted(other_names)),
        sequence=tuple(sequence) if sequence else None,
    )


@lru_cache(maxsize=_LRU_CACHE_SIZE)
def __is_rule_option_put_before(
    rule: Rule,
    name: str,
    other_names: Union[Sequence[str], set[str]],
    sequence: Optional[tuple[str, ...]] = None,
) -> Optional[bool]:
    if len(other_names) == 0:
        _logger.debug(
            "Cannot unambiguously determine if option %s is put before empty Iterable of other options.",
            name,
        )
        return None

    positions = get_rule_option_positions(rule, name, sequence=sequence)

    if name in other_names:
        _logger.debug("Excluding name %s from other_names because of overlap.", name)
        other_names = set(other_names).difference({name})

    other_positions = get_rule_options_positions(rule, other_names, sequence=sequence)

    for other_position in other_positions:
        for position in positions:
            if position < other_position:
                return True
    return False


def is_rule_option_always_put_before(
    rule: Rule,
    name: str,
    other_names: Union[Sequence[str], set[str]],
    sequence: Optional[Iterable[str]] = None,
) -> Optional[bool]:
    """Checks whether a rule option is placed before one or more other options."""
    return __is_rule_option_always_put_before(
        rule,
        name,
        tuple(sorted(other_names)),
        sequence=tuple(sequence) if sequence else None,
    )


@lru_cache(maxsize=_LRU_CACHE_SIZE)
def __is_rule_option_always_put_before(
    rule: Rule,
    name: str,
    other_names: Union[Sequence[str], set[str]],
    sequence: Optional[tuple[str, ...]] = None,
) -> Optional[bool]:
    if len(other_names) == 0:
        _logger.debug(
            "Cannot unambiguously determine if option %s is put before empty Iterable of other options.",
            name,
        )
        return None

    positions = get_rule_option_positions(rule, name, sequence=sequence)

    if name in other_names:
        _logger.debug("Excluding name %s from other_names because of overlap.", name)
        other_names = set(other_names).difference({name})

    other_positions = get_rule_options_positions(rule, other_names, sequence=sequence)

    for other_position in other_positions:
        for position in positions:
            if position >= other_position:
                return False
    return True


def are_rule_options_put_before(
    rule: Rule,
    names: Union[Sequence[str], set[str]],
    other_names: Union[Sequence[str], set[str]],
    sequence: Optional[Iterable[str]] = None,
) -> Optional[bool]:
    """Checks whether rule options are placed before one or more other options."""
    if len(other_names) == 0:
        _logger.debug(
            "Cannot unambiguously determine if an empty Iterable of options are put before other options %s.",
            other_names,
        )
        return None
    if len(other_names) == 0:
        _logger.debug(
            "Cannot unambiguously determine if options %s are put before empty Iterable of other options.",
            names,
        )
        return None

    for name in names:
        if is_rule_option_put_before(rule, name, other_names, sequence=sequence):
            return True
    return False


def are_rule_options_always_put_before(
    rule: Rule,
    names: Iterable[str],
    other_names: Sequence[str],
    sequence: Optional[Iterable[str]] = None,
) -> Optional[bool]:
    """Checks whether rule options are placed before one or more other options."""
    if len(other_names) == 0:
        _logger.debug(
            "Cannot unambiguously determine if an empty Iterable of options are put before other options %s.",
            other_names,
        )
        return None
    if len(other_names) == 0:
        _logger.debug(
            "Cannot unambiguously determine if options %s are put before empty Iterable of other options.",
            names,
        )
        return None

    for name in names:
        if not is_rule_option_put_before(rule, name, other_names, sequence=sequence):
            return False
    return True


def select_rule_options_by_regex(rule: Rule, regex) -> Iterable[str]:  # noqa: ANN001
    """Selects rule options present in rule matching a regular expression."""
    return __select_rule_options_by_regex(rule, regex)


@lru_cache(maxsize=_LRU_CACHE_SIZE)
def __select_rule_options_by_regex(rule: Rule, regex) -> Iterable[str]:  # noqa: ANN001
    options = []

    for option in rule["options"]:
        name = option["name"]
        if _regex_provider.match(regex, name):
            options.append(name)

    return tuple(sorted(options))


def get_rule_keyword_sequences(
    rule: Rule,
    seperator_keywords: Iterable[str] = BUFFER_KEYWORDS,
    included_keywords: Iterable[str] = ALL_DETECTION_KEYWORDS,
) -> Sequence[tuple[str, ...]]:
    """Returns a sequence of sequences of detection options in a rule."""
    sequences: list[list[str]] = []

    # Relies on the assumption that the order of options in the rule is preserved while parsing
    sequence_i = -1
    first_seperator_seen = False
    for option in rule["options"]:
        name = option["name"]
        if name in seperator_keywords:
            if not first_seperator_seen:
                if len(sequences) > 0:
                    sequences[sequence_i].append(name)
                else:
                    sequence_i += 1
                    sequences.append([name])
            else:
                sequence_i += 1
                sequences.append([name])
            first_seperator_seen = True
        elif name in included_keywords and sequence_i == -1:
            sequence_i += 1
            sequences.append([name])
        elif name in included_keywords:
            sequences[sequence_i].append(name)

    if len(sequences) == 0:
        _logger.debug(
            "No sequences found separated by %s in rule %s",
            seperator_keywords,
            rule["raw"],
        )
        return ()

    for sequence in sequences:
        assert len(sequence) > 0

    result = tuple(tuple(sequence) for sequence in sequences)

    _logger.debug(
        "Detected sequences %s separated by %s in rule %s",
        result,
        seperator_keywords,
        rule["raw"],
    )

    return result



"""`UnexpectedChecker`."""

import logging

from suricata_check.checkers.interface import CheckerInterface
from suricata_check.utils.checker import (
    is_rule_option_set,
    is_rule_suboption_set,
)
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue, Rule


class UnexpectedChecker(CheckerInterface):
    """The `UnexpectedChecker` contains several checks for unexpected Suricata behavior that users may not anticipate.

    Codes C000-C010 report on unexpected behavior.
    """

    codes = {
        "C000": {"severity": logging.WARNING},
    }

    def _check_rule(
        self: "UnexpectedChecker",
        rule: Rule,
    ) -> ISSUES_TYPE:
        issues: ISSUES_TYPE = []

        if (
            is_rule_suboption_set(rule, "flowbits", "set")
            or is_rule_suboption_set(rule, "xbits", "set")
        ) and (is_rule_option_set(rule, "threshold")):
            issues.append(
                Issue(
                    code="C000",
                    message="""\
The rule uses the Suricata threshold option in combination with the setting of flowbits or xbits.
Note that the flowbit or xbit will be set on every match regardless of whether the threshold is reached.
Consider removing the `threshold` option from the rule to prevent confusion.\
""",
                ),
            )

        return issues



"""`MetadataChecker`."""

import logging

from suricata_check.checkers.interface import CheckerInterface
from suricata_check.utils.checker import (
    is_rule_option_set,
    is_rule_suboption_set,
)
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue, Rule


class MetadataChecker(CheckerInterface):
    """The `MetadataChecker` contains several checks for Suricata metadata options.

    Codes S800-810 report on missing common `metadata` fields
    """

    codes = {
        "S800": {"severity": logging.INFO},
        "S801": {"severity": logging.INFO},
        "S802": {"severity": logging.INFO},
        "S803": {"severity": logging.INFO},
    }

    def _check_rule(
        self: "MetadataChecker",
        rule: Rule,
    ) -> ISSUES_TYPE:
        issues: ISSUES_TYPE = []

        if not is_rule_suboption_set(rule, "metadata", "attack_target"):
            issues.append(
                Issue(
                    code="S800",
                    message="""\
The rule did not specify the `attack_target` metadata option.
Consider specifying the `attack_target` metadata option to help analysts interpret alerts raised by this rule.\
""",
                ),
            )

        if not is_rule_suboption_set(rule, "metadata", "signature_severity") and not (
            is_rule_option_set(rule, "noalert")
            or is_rule_suboption_set(rule, "flowbits", "noalert")
        ):
            issues.append(
                Issue(
                    code="S801",
                    message="""\
The rule did not specify the `signature_severity` metadata option.
Consider specifying the `signature_severity` metadata option to help analysts interpret alerts raised by this rule.\
""",
                ),
            )

        if not is_rule_suboption_set(rule, "metadata", "performance_impact"):
            issues.append(
                Issue(
                    code="S802",
                    message="""\
The rule did not specify the `performance_impact` metadata option.
Consider specifying the `performance_impact` metadata option to help SOCs determine when to enable this rule.\
""",
                ),
            )

        if not is_rule_suboption_set(rule, "metadata", "deployment"):
            issues.append(
                Issue(
                    code="S803",
                    message="""\
The rule did not specify the `deployment` metadata option. \
Consider specifying the `deployment` metadata option to help SOCs determine when to enable this rule.\
""",
                ),
            )

        return issues



import io
import json
import logging
import os
from collections import defaultdict
from collections.abc import Mapping, Sequence
from typing import Optional, Union

import click
import tabulate

from suricata_check._checkers import get_checkers
from suricata_check.checkers.interface.checker import CheckerInterface
from suricata_check.utils.checker_typing import (
    EXTENSIVE_SUMMARY_TYPE,
    ISSUES_TYPE,
    RULE_REPORTS_TYPE,
    RULE_SUMMARY_TYPE,
    SIMPLE_SUMMARY_TYPE,
    OutputReport,
    OutputSummary,
    Rule,
    RuleReport,
)

GITLAB_SEVERITIES = {
    logging.DEBUG: "info",
    logging.INFO: "info",
    logging.WARNING: "minor",
    logging.ERROR: "major",
    logging.CRITICAL: "critical",
}
GITHUB_SEVERITIES = {
    logging.DEBUG: "debug",
    logging.INFO: "notice",
    logging.WARNING: "warning",
    logging.ERROR: "error",
    logging.CRITICAL: "error",
}
GITHUB_COMMAND = (
    "::{level} file={file},line={line},endLine={end_line},title={title}::{message}"
)

_logger = logging.getLogger(__name__)


def write_output(
    output: OutputReport,
    out: str,
    gitlab: bool = False,
    github: bool = False,
    rules_file: Optional[str] = None,
) -> None:
    _logger.info(
        "Writing output to suricata-check.jsonl and suricata-check-fast.log in %s",
        os.path.abspath(out),
    )
    with (
        open(
            os.path.join(out, "suricata-check.jsonl"),
            "w",
            buffering=io.DEFAULT_BUFFER_SIZE,
        ) as jsonl_fh,
        open(
            os.path.join(out, "suricata-check-fast.log"),
            "w",
            buffering=io.DEFAULT_BUFFER_SIZE,
        ) as fast_fh,
    ):
        rules: RULE_REPORTS_TYPE = output.rules
        jsonl_fh.write("\n".join([str(rule) for rule in rules]))

        for rule_report in rules:
            rule: Rule = rule_report.rule
            lines: str = (
                "{}-{}".format(rule_report.line_begin, rule_report.line_end)
                if rule_report.line_begin
                else "Unknown"
            )
            issues: ISSUES_TYPE = rule_report.issues
            for issue in issues:
                code = issue.code
                severity = (
                    logging.getLevelName(issue.severity) if issue.severity else None
                )
                issue_msg = issue.message.replace("\n", " ")

                msg = "[{}]{} Lines {}, sid {}: {}".format(
                    code,
                    f" ({severity})" if severity else "",
                    lines,
                    rule["sid"],
                    issue_msg,
                )
                fast_fh.write(msg + "\n")
                click.secho(msg, color=True, fg="blue")

    if output.summary is not None:
        __write_output_stats(output, out)

    if gitlab:
        assert rules_file is not None

        __write_output_gitlab(output, out, rules_file)

    if github:
        assert rules_file is not None

        __write_output_github(output, rules_file)


def __write_output_stats(output: OutputReport, out: str) -> None:
    assert output.summary is not None

    with open(
        os.path.join(out, "suricata-check-stats.log"),
        "w",
        buffering=io.DEFAULT_BUFFER_SIZE,
    ) as stats_fh:
        summary: OutputSummary = output.summary

        overall_summary: SIMPLE_SUMMARY_TYPE = summary.overall_summary

        n_issues = overall_summary["Total Issues"]
        n_rules = (
            overall_summary["Rules with Issues"]
            + overall_summary["Rules without Issues"]
        )

        stats_fh.write(
            tabulate.tabulate(
                (
                    (
                        k,
                        v,
                        (
                            "{:.0%}".format(v / n_rules)
                            if k.startswith("Rules ") and n_rules > 0
                            else "-"
                        ),
                    )
                    for k, v in overall_summary.items()
                ),
                headers=(
                    "Count",
                    "Percentage of Rules",
                ),
            )
            + "\n\n",
        )

        click.secho(
            f"Total issues found: {overall_summary['Total Issues']}",
            color=True,
            bold=True,
            fg="blue",
        )
        click.secho(
            f"Rules with Issues found: {overall_summary['Rules with Issues']}",
            color=True,
            bold=True,
            fg="blue",
        )

        issues_by_group: SIMPLE_SUMMARY_TYPE = summary.issues_by_group

        stats_fh.write(
            tabulate.tabulate(
                (
                    (k, v, "{:.0%}".format(v / n_issues) if n_issues > 0 else "-")
                    for k, v in issues_by_group.items()
                ),
                headers=(
                    "Count",
                    "Percentage of Total Issues",
                ),
            )
            + "\n\n",
        )

        issues_by_type: EXTENSIVE_SUMMARY_TYPE = summary.issues_by_type
        for checker, checker_issues_by_type in issues_by_type.items():
            stats_fh.write(" " + checker + " " + "\n")
            stats_fh.write("-" * (len(checker) + 2) + "\n")
            stats_fh.write(
                tabulate.tabulate(
                    (
                        (
                            k,
                            v,
                            "{:.0%}".format(v / n_rules) if n_rules > 0 else "-",
                        )
                        for k, v in checker_issues_by_type.items()
                    ),
                    headers=(
                        "Count",
                        "Percentage of Rules",
                    ),
                )
                + "\n\n",
            )


def __write_output_gitlab(output: OutputReport, out: str, rules_file: str) -> None:
    with open(
        os.path.join(out, "suricata-check-gitlab.json"),
        "w",
        buffering=io.DEFAULT_BUFFER_SIZE,
    ) as gitlab_fh:
        issue_dicts = []
        for rule_report in output.rules:
            line_begin: Optional[int] = rule_report.line_begin
            assert line_begin is not None
            line_end: Optional[int] = rule_report.line_end
            assert line_end is not None
            issues: ISSUES_TYPE = rule_report.issues
            for issue in issues:
                code = issue.code
                issue_msg = issue.message.replace("\n", " ")
                assert issue.checker is not None
                issue_checker = issue.checker
                issue_hash = str(issue.hash)
                assert issue.severity is not None
                issue_severity = GITLAB_SEVERITIES[issue.severity]

                issue_dict: Mapping[
                    str,
                    Union[str, list[str], Mapping[str, Union[str, Mapping[str, int]]]],
                ] = {
                    "description": issue_msg,
                    "categories": [issue_checker],
                    "check_name": f"Suricata Check {code}",
                    "fingerprint": issue_hash,
                    "severity": issue_severity,
                    "location": {
                        "path": rules_file,
                        "lines": {"begin": line_begin, "end": line_end},
                    },
                }
                issue_dicts.append(issue_dict)

        gitlab_fh.write(json.dumps(issue_dicts))


def __write_output_github(output: OutputReport, rules_file: str) -> None:
    output_lines: dict[str, list[str]] = {
        k: [] for k in set(GITHUB_SEVERITIES.values())
    }
    for rule_report in output.rules:
        line_begin: Optional[int] = rule_report.line_begin
        assert line_begin is not None
        line_end: Optional[int] = rule_report.line_end
        assert line_end is not None
        issues: ISSUES_TYPE = rule_report.issues
        for issue in issues:
            code = issue.code
            issue_msg = issue.message.replace("\n", " ")
            assert issue.checker is not None
            issue_checker = issue.checker
            assert issue.severity is not None
            issue_severity = GITHUB_SEVERITIES[issue.severity]
            title = f"{issue_checker} - {code}"

            output_lines[issue_severity].append(
                GITHUB_COMMAND.format(
                    level=issue_severity,
                    file=rules_file,
                    line=line_begin,
                    end_line=line_end,
                    title=title,
                    message=issue_msg,
                )
            )

    for message_level, lines in output_lines.items():
        if len(lines) > 0:
            print(f"::group::{message_level}")  # noqa: T201
            for message in lines:
                print(message)  # noqa: T201
            print("::endgroup::")  # noqa: T201


def summarize_rule(
    rule: RuleReport,
    checkers: Optional[Sequence[CheckerInterface]] = None,
) -> RULE_SUMMARY_TYPE:
    """Summarizes the issues found in a rule.

    Args:
    rule: The rule output dictionary to be summarized.
    checkers: The checkers to be used to check the rule.

    Returns:
    A dictionary containing a summary of all issues found in the rule.

    """
    if checkers is None:
        checkers = get_checkers()

    summary = {}

    issues: ISSUES_TYPE = rule.issues
    summary["total_issues"] = len(issues)
    summary["issues_by_group"] = defaultdict(int)
    for issue in issues:
        checker = issue.checker
        summary["issues_by_group"][checker] += 1

    # Ensure also checkers without issues are included in the report.
    for checker in checkers:
        if checker.__class__.__name__ not in summary["issues_by_group"]:
            summary["issues_by_group"][checker.__class__.__name__] = 0

    # Sort dictionaries for deterministic output
    summary["issues_by_group"] = __sort_mapping(summary["issues_by_group"])

    return summary


def summarize_output(
    output: OutputReport,
    checkers: Optional[Sequence[CheckerInterface]] = None,
) -> OutputSummary:
    """Summarizes the issues found in a rules file.

    Args:
    output: The unsammarized output of the rules file containing all rules and their issues.
    checkers: The checkers to be used to check the rule.

    Returns:
    A dictionary containing a summary of all issues found in the rules file.

    """
    if checkers is None:
        checkers = get_checkers()

    return OutputSummary(
        overall_summary=__get_overall_summary(output),
        issues_by_group=__get_issues_by_group(output, checkers),
        issues_by_type=__get_issues_by_type(output, checkers),
    )


def __get_overall_summary(
    output: OutputReport,
) -> SIMPLE_SUMMARY_TYPE:
    overall_summary = {
        "Total Issues": 0,
        "Rules with Issues": 0,
        "Rules without Issues": 0,
    }

    rules: RULE_REPORTS_TYPE = output.rules
    for rule in rules:
        issues: ISSUES_TYPE = rule.issues
        overall_summary["Total Issues"] += len(issues)

        if len(issues) == 0:
            overall_summary["Rules without Issues"] += 1
        else:
            overall_summary["Rules with Issues"] += 1

    return overall_summary


def __get_issues_by_group(
    output: OutputReport,
    checkers: Optional[Sequence[CheckerInterface]] = None,
) -> SIMPLE_SUMMARY_TYPE:
    if checkers is None:
        checkers = get_checkers()

    issues_by_group = defaultdict(int)

    # Ensure also checkers and codes without issues are included in the report.
    for checker in checkers:
        issues_by_group[checker.__class__.__name__] = 0

    rules: RULE_REPORTS_TYPE = output.rules
    for rule in rules:
        issues: ISSUES_TYPE = rule.issues

        for issue in issues:
            checker = issue.checker
            if checker is not None:
                issues_by_group[checker] += 1

    return __sort_mapping(issues_by_group)


def __get_issues_by_type(
    output: OutputReport,
    checkers: Optional[Sequence[CheckerInterface]] = None,
) -> EXTENSIVE_SUMMARY_TYPE:
    if checkers is None:
        checkers = get_checkers()
    issues_by_type: EXTENSIVE_SUMMARY_TYPE = defaultdict(lambda: defaultdict(int))

    # Ensure also checkers and codes without issues are included in the report.
    for checker in checkers:
        for code in checker.codes:
            issues_by_type[checker.__class__.__name__][code] = 0

    rules: RULE_REPORTS_TYPE = output.rules
    for rule in rules:
        issues: ISSUES_TYPE = rule.issues

        checker_codes = defaultdict(lambda: defaultdict(int))
        for issue in issues:
            checker = issue.checker
            if checker is not None:
                code = issue.code
                checker_codes[checker][code] += 1

        for checker, codes in checker_codes.items():
            for code, count in codes.items():
                issues_by_type[checker][code] += count

    for key in issues_by_type:
        issues_by_type[key] = __sort_mapping(issues_by_type[key])

    return __sort_mapping(issues_by_type)


def __sort_mapping(mapping: Mapping) -> dict:
    return {key: mapping[key] for key in sorted(mapping.keys())}



"""The `suricata_check.checkers.community` modules contains several checkers based on community issues, such as this GitHub.

Reference: TODO
"""

from suricata_check.checkers.community.best import BestChecker
from suricata_check.checkers.community.unexpected import UnexpectedChecker

__all__ = [
    "BestChecker",
    "UnexpectedChecker",
]



"""The `suricata_check.suricata_check` module contains the command line utility and the main program logic."""

import logging
import logging.handlers
import pkgutil
from collections.abc import Sequence
from functools import lru_cache

from suricata_check.checkers.interface import CheckerInterface
from suricata_check.checkers.interface.dummy import DummyChecker
from suricata_check.utils.checker_typing import (
    get_all_subclasses,
)
from suricata_check.utils.regex import get_regex_provider

_regex_provider = get_regex_provider()

_logger = logging.getLogger(__name__)

# Global variable to check if extensions have already been imported in case get_checkers() is called multiple times.
suricata_check_extensions_imported = False


def _import_extensions() -> None:
    global suricata_check_extensions_imported  # noqa: PLW0603
    if suricata_check_extensions_imported is True:
        return

    for module in pkgutil.iter_modules():
        if module.name.startswith("suricata_check_"):
            try:
                imported_module = __import__(module.name)
                _logger.info(
                    "Detected and successfully imported suricata-check extension %s with version %s.",
                    module.name.replace("_", "-"),
                    getattr(imported_module, "__version__"),
                )
            except ImportError:
                _logger.warning(
                    "Detected potential suricata-check extension %s but failed to import it.",
                    module.name.replace("_", "-"),
                )
    suricata_check_extensions_imported = True


@lru_cache(maxsize=1)
def get_checkers(
    include: Sequence[str] = (".*",),
    exclude: Sequence[str] = (),
    issue_severity: int = logging.INFO,
) -> Sequence[CheckerInterface]:
    """Auto discovers all available checkers that implement the CheckerInterface.

    Returns:
    A list of available checkers that implement the CheckerInterface.

    """
    # Check for extensions and try to import them
    _import_extensions()

    checkers: list[CheckerInterface] = []
    for checker in get_all_subclasses(CheckerInterface):
        if checker.__name__ == DummyChecker.__name__:
            continue

        # Initialize DummyCheckers to retrieve error messages.
        if issubclass(checker, DummyChecker):
            checker()

        enabled, relevant_codes = __get_checker_enabled(
            checker, include, exclude, issue_severity
        )

        if enabled:
            checkers.append(checker(include=relevant_codes))

        else:
            _logger.info("Checker %s is disabled.", checker.__name__)

    _logger.info(
        "Discovered and enabled checkers: [%s]",
        ", ".join([c.__class__.__name__ for c in checkers]),
    )
    if len(checkers) == 0:
        _logger.warning(
            "No checkers were enabled. Check the include and exclude arguments."
        )

    # Perform a uniqueness check on the codes emmitted by the checkers
    for checker1 in checkers:
        for checker2 in checkers:
            if checker1 == checker2:
                continue
            if not set(checker1.codes).isdisjoint(checker2.codes):
                msg = f"Checker {checker1.__class__.__name__} and {checker2.__class__.__name__} have overlapping codes."
                _logger.error(msg)

    return sorted(checkers, key=lambda x: x.__class__.__name__)


def __get_checker_enabled(
    checker: type[CheckerInterface],
    include: Sequence[str],
    exclude: Sequence[str],
    issue_severity: int,
) -> tuple[bool, set[str]]:
    enabled = checker.enabled_by_default

    # If no include regexes are provided, include all by default
    if len(include) == 0:
        relevant_codes = set(checker.codes.keys())
    else:
        # If include regexes are provided, include all codes that match any of these regexes
        relevant_codes = set()

        for regex in include:
            relevant_codes.update(
                set(
                    filter(
                        lambda code: _regex_provider.compile("^" + regex + "$").match(
                            code
                        )
                        is not None,
                        checker.codes.keys(),
                    )
                )
            )

        if len(relevant_codes) > 0:
            enabled = True

    # Now remove the codes that are excluded according to any of the provided exclude regexes
    for regex in exclude:
        relevant_codes = set(
            filter(
                lambda code: _regex_provider.compile("^" + regex + "$").match(code)
                is None,
                relevant_codes,
            )
        )

    # Now filter out irrelevant codes based on severity
    relevant_codes = set(
        filter(
            lambda code: checker.codes[code]["severity"] >= issue_severity,
            relevant_codes,
        )
    )

    if len(relevant_codes) == 0:
        enabled = False

    return enabled, relevant_codes



"""`PcreChecker`."""

import logging

from suricata_check.checkers.interface import CheckerInterface
from suricata_check.utils.checker import (
    is_rule_option_equal_to_regex,
    is_rule_option_set,
)
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue, Rule
from suricata_check.utils.regex import get_regex_provider

_regex_provider = get_regex_provider()

_S601_REGEX = _regex_provider.compile(
    r"^.*(\.\*).*$",
    _regex_provider.IGNORECASE,
)


class PcreChecker(CheckerInterface):
    """The `PcreChecker` contains several checks for Suricata PCRE options.

    Codes S600-610 report on unrecommended usages of `pcre`
    """

    codes = {"S600": {"severity": logging.INFO}, "S601": {"severity": logging.INFO}}

    def _check_rule(
        self: "PcreChecker",
        rule: Rule,
    ) -> ISSUES_TYPE:
        issues: ISSUES_TYPE = []

        if is_rule_option_set(rule, "pcre") and not is_rule_option_set(rule, "content"):
            issues.append(
                Issue(
                    code="S600",
                    message="""\
The rule uses the `pcre` option but has no `content` option set.
Consider using the content option atleast once to anchor and improve runtime performance.\
""",
                ),
            )

        if is_rule_option_set(rule, "pcre") and is_rule_option_equal_to_regex(
            rule, "pcre", _S601_REGEX
        ):
            issues.append(
                Issue(
                    code="S601",
                    message="""\
The rule uses the `pcre` with an unlimited inspection depth.
Consider limiting the inspection depth to improve runtime performance.\
""",
                ),
            )

        return issues



"""The `suricata_check.utils` module contains several utilities for the suricata-check main program and the checkers."""

from suricata_check.utils import checker, checker_typing, regex
from suricata_check.utils._click import ClickHandler
from suricata_check.utils._path import find_rules_file

__all__ = (
    "ClickHandler",
    "checker",
    "checker_typing",
    "find_rules_file",
    "regex",
)



"""`PerformanceChecker`."""

import logging

from suricata_check.checkers.interface import CheckerInterface
from suricata_check.utils.checker import (
    get_rule_keyword_sequences,
    is_rule_option_set,
)
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue, Rule
from suricata_check.utils.regex import (
    BASE64_BUFFER_KEYWORDS,
    BASE64_TRANSFORMATION_KEYWORDS,
)


class PerformanceChecker(CheckerInterface):
    """The `PerformanceChecker` contains several checks for Suricata performance issues.

    Codes S900-910 report on usage of options that can slow the detection engine.
    """

    codes = {
        "S900": {"severity": logging.INFO},
        "S901": {"severity": logging.INFO},
        "S902": {"severity": logging.INFO},
        "S903": {"severity": logging.INFO},
    }

    def _check_rule(
        self: "PerformanceChecker",
        rule: Rule,
    ) -> ISSUES_TYPE:
        issues: ISSUES_TYPE = []

        if is_rule_option_set(rule, "http.response_body"):
            issues.append(
                Issue(
                    code="S900",
                    message="""\
The rule uses the `http.response_body` option, which is known to be slow in Suricata 5.
Consider specifying the `file.data` option instead.\
""",
                ),
            )

        for option in BASE64_BUFFER_KEYWORDS + BASE64_TRANSFORMATION_KEYWORDS:
            if is_rule_option_set(rule, option):
                issues.append(
                    Issue(
                        code="S901",
                        message="""\
The rule uses a `base64_` keyword, which is known to be slow.
Consider detection methods avoiding the usage of `base64_` keywords to improve runtime performance.\
""",
                    ),
                )

        for sequence in get_rule_keyword_sequences(rule):
            if "http.uri" in sequence and "bsize" in sequence:
                issues.append(
                    Issue(
                        code="S902",
                        message="""\
The rule uses the `bsize` keyword on the `http.uri` buffer, which is known to be slow.
Consider using the `urilen` option instead to improve runtime performance.\
""",
                    ),
                )

        return issues



"""The `suricata_check.checkers.interface.checker` module contains the `CheckerInterface`.

Implementation of the `CheckerInterface` is neccessary for checker auto-discovery.
"""

import abc
import logging
from collections.abc import Iterable, Mapping
from typing import Optional

from suricata_check.utils.checker import get_rule_option, is_rule_option_set
from suricata_check.utils.checker_typing import ISSUES_TYPE, Rule

_logger = logging.getLogger(__name__)


class CheckerInterface:
    """Interface for rule checkers returning a list of issues.

    These checkers are automatically discovered through `suricata_check.suricata_check.get_checkers()`.

    Each code should start with an upper case letter (may be multiple), followed by three digits.
    In other words, each code should follow the following regex `[A-Z]{1,}[0-9]{3}`

    We recommend using a letter to indicate the category of the issue, such as described in `README.md`.
    Moreover, we suggest to reserve certain ranges of numbers for each checker.

    """

    codes: Mapping[str, Mapping[str, int]]
    """A Mapping of issue codes emitted by the checker to metadata for those issue types.
    The metadata is structured in the form of a Mapping from attribute name to attribute value.
    The one mandatory metadata attribute is severity, which must be one of the levels provided by the `logging` module"""

    enabled_by_default: bool = True
    """A boolean indicating if the checker is enabled by default when discovered automatically."""

    def __init__(
        self: "CheckerInterface", include: Optional[Iterable[str]] = None
    ) -> None:
        """Initializes the checker given a list of issue codes to emit."""
        if include is None:
            include = self.codes
        self.include = include

        super().__init__()

    def check_rule(
        self: "CheckerInterface",
        rule: Rule,
    ) -> ISSUES_TYPE:
        """Checks a rule and returns a list of issues found."""
        self.__log_rule_processing(rule)
        return self.__add_checker_metadata(
            self.__add_issue_metadata(self.__filter_issues(self._check_rule(rule)))
        )

    @abc.abstractmethod
    def _check_rule(
        self: "CheckerInterface",
        rule: Rule,
    ) -> ISSUES_TYPE:
        """Checks a rule and returns a list of issues found."""

    def __log_rule_processing(
        self: "CheckerInterface",
        rule: Rule,
    ) -> None:
        sid: Optional[int] = None
        if is_rule_option_set(rule, "sid"):
            sid_str = get_rule_option(rule, "sid")
            assert sid_str is not None
            sid = int(sid_str)

        _logger.debug("Running %s on rule %s", self.__class__.__name__, sid)

    def __add_issue_metadata(
        self: "CheckerInterface",
        issues: ISSUES_TYPE,
    ) -> ISSUES_TYPE:
        """Given a list of issues, return the same list with metadata from the issue types."""
        for issue in issues:
            metadata = self.codes[issue.code]
            if "severity" in metadata:
                issue.severity = metadata["severity"]

        return issues

    def __add_checker_metadata(
        self: "CheckerInterface",
        issues: ISSUES_TYPE,
    ) -> ISSUES_TYPE:
        """Given a list of issues, return the same list with metadata from the checker."""
        name = self.__class__.__name__

        for issue in issues:
            issue.checker = name

        return issues

    def __filter_issues(
        self: "CheckerInterface",
        issues: ISSUES_TYPE,
    ) -> ISSUES_TYPE:
        """Given a list of issues, return the same list having filtered out disabled issue types."""
        filtered_issues = []

        for issue in issues:
            if issue.code in self.include:
                filtered_issues.append(issue)
            elif issue.code not in self.codes:
                _logger.warning(
                    "Issue with filtered code %s not found in checker %s",
                    issue.code,
                    self.__class__.__name__,
                )

        return filtered_issues



"""`SidChecker`."""

import logging
from collections.abc import Mapping, Sequence
from typing import Optional

from suricata_check.checkers.interface import CheckerInterface
from suricata_check.utils.checker import get_rule_option
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue, Rule
from suricata_check.utils.regex import get_regex_provider

SID_ALLOCATION: Mapping[str, Sequence[tuple[int, int]]] = {
    "local": [(1000000, 1999999)],
    "ET OPEN": [
        (2000000, 2103999),
        (2400000, 2609999),
    ],
    "ET": [(2700000, 2799999)],
    "ETPRO": [(2800000, 2899999)],
}

_regex_provider = get_regex_provider()

_MSG_PREFIX_REGEX = _regex_provider.compile(r"^\"([A-Z0-9 ]*).*\"$")

_logger = logging.getLogger(__name__)


class SidChecker(CheckerInterface):
    """The `SidChecker` contains several checks based on the Suricata SID allocation.

    Specifically, the `SidChecker` checks for the following:
        S300: Allocation to reserved SID range, whereas no range is reserved for the rule.

        S301: Allocation to unallocated SID range, whereas local range should be used.

        S302: Allocation to wrong reserved SID range, whereas another reserved range should be used.

        S303: Allocation to unallocated SID range, whereas a reserved range should be used.
    """

    codes = {
        "S300": {"severity": logging.INFO},
        "S301": {"severity": logging.INFO},
        "S302": {"severity": logging.INFO},
        "S303": {"severity": logging.INFO},
    }

    def _check_rule(
        self: "SidChecker",
        rule: Rule,
    ) -> ISSUES_TYPE:
        issues: ISSUES_TYPE = []

        sid = get_rule_option(rule, "sid")
        msg = get_rule_option(rule, "msg")

        if sid is not None and msg is not None:
            sid = int(sid)
            range_name = self.__get_range_name(sid, SID_ALLOCATION)
            prefix = self.__get_msg_prefix(msg)

            if (
                prefix not in SID_ALLOCATION.keys()
                and range_name is not None
                and range_name != "local"
            ):
                issues.append(
                    Issue(
                        code="S300",
                        message=f"""\
Allocation to reserved SID range, whereas no range is reserved for the rule.
Consider using an sid in one of the following ranges: {SID_ALLOCATION["local"]}.\
""",
                    ),
                )

            if prefix not in SID_ALLOCATION.keys() and range_name is None:
                issues.append(
                    Issue(
                        code="S301",
                        message=f"""\
Allocation to unallocated SID range, whereas local range should be used.
Consider using an sid in one of the following ranges: {SID_ALLOCATION["local"]}.\
""",
                    ),
                )

            if prefix in SID_ALLOCATION.keys() and (
                range_name is not None
                and not (prefix + " ").startswith(range_name + " ")
                and not (range_name + " ").startswith(prefix + " ")
            ):
                issues.append(
                    Issue(
                        code="S302",
                        message=f"""\
Allocation to wrong reserved SID range, whereas another reserved range should be used.
Consider using an sid in one of the following ranges: {SID_ALLOCATION[prefix]}.\
""",
                    ),
                )

            if prefix in SID_ALLOCATION.keys() and range_name is None:
                issues.append(
                    Issue(
                        code="S303",
                        message=f"""\
Allocation to unallocated SID range, whereas a reserved range should be used.
Consider using an sid in one of the following ranges: {SID_ALLOCATION[prefix]}.\
""",
                    ),
                )

        return issues

    @staticmethod
    def __in_range(sid: int, sid_range: Sequence[tuple[int, int]]) -> bool:
        for start, end in sid_range:
            if start <= sid <= end:
                return True

        return False

    @staticmethod
    def __get_range_name(
        sid: int,
        ranges: Mapping[str, Sequence[tuple[int, int]]],
    ) -> Optional[str]:
        for range_name, sid_range in ranges.items():
            for start, end in sid_range:
                if start <= sid <= end:
                    _logger.debug("Detected sid from range: %s", range_name)
                    return range_name
        return None

    @staticmethod
    def __get_msg_prefix(msg: str) -> str:
        match = _MSG_PREFIX_REGEX.match(msg)
        assert match is not None

        parts = match.group(1).strip().split(" ")
        prefix: str = ""
        for i in list(reversed(range(len(parts)))):
            prefix = " ".join(parts[: i + 1])
            if prefix in SID_ALLOCATION.keys() or " " not in prefix:
                break

        _logger.debug("Detected prefix: %s", prefix)

        return prefix



"""Wrapper around idstools.rule.Rule for future removal of idstools dependency."""

from typing import Optional

import idstools.rule


class Rule:
    """Wrapper around an `suricata_check.rule.Rule` in preparation for dropping the idstools dependency."""

    def __init__(self, inner: Optional[idstools.rule.Rule]) -> None:
        """Create a wrapper around an existing `suricata_check.rule.Rule`.

        Parameters
        ----------
        inner:
            The parsed rule object from `suricata_check.rule.parse()` or None.
        """
        super().__init__()
        self._inner = inner

    @property
    def inner(self) -> Optional[idstools.rule.Rule]:
        """Return the underlying parsed rule object (or None)."""
        return self._inner

    def __getitem__(self, key: str):  # noqa: ANN204
        """Forward mapping access to the underlying rule.

        Raises KeyError when no underlying rule is present.
        """
        if self._inner is None:
            raise KeyError(key)
        return self._inner[key]

    def get(self, key: str, default: Optional[object] = None):  # noqa: ANN201
        """Return the value for *key* if present, otherwise *default*."""
        if self._inner is None:
            return default
        return self._inner.get(key, default)

    def __contains__(self, key: str) -> bool:
        """Return True when the underlying rule contains *key*."""
        return self._inner is not None and key in self._inner

    def __repr__(self) -> str:
        """Return representation of the wrapped rule (or 'None')."""
        return repr(self._inner)


class ParsingError(RuntimeError):
    """Raised when a rule cannot be parsed by suricata-check.

    Most likely, such a rule is also an invalid Suricata rule.
    """

    def __init__(self: "ParsingError", message: str) -> None:
        """Initializes the `ParsingError` with the raw rule as message."""
        super().__init__(message)


def parse(text: Optional[str]) -> Optional["Rule"]:
    """Parse a rule string using the underlying `idstools` parser.

    Return a wrapped `Rule` instance.

    Returns None when the text could not be parsed as a rule.
    """
    try:
        inner = idstools.rule.parse(text)
    except Exception as e:  # noqa: BLE001
        raise ParsingError(str(e))

    if inner is None:
        return None

    return Rule(inner)



"""The `suricata_check.checkers` module contains all rule checkers."""

# Exception may occur if pytest is not installed.
try:
    from suricata_check.tests.checker import GenericChecker

    __all__ = [
        "GenericChecker",
    ]
except ModuleNotFoundError:
    import sys as _sys

    # Only warn about failed import if not used as CLI.
    if not _sys.argv[0].endswith("suricata_check.py"):
        import logging as _logging

        _logger = _logging.getLogger(__name__)
        _logger.warning(
            """Failed to initialize `suricata_check.tests.checker.GenericChecker` due to failed imports. \
    Ensure all necessary development dependencies are installed if you need to run tests.""",
        )
    __all__ = []



"""`MsgChecker`."""

import logging

from suricata_check.checkers.interface import CheckerInterface
from suricata_check.utils.checker import (
    get_rule_option,
    is_rule_option_equal_to_regex,
    is_rule_option_set,
    is_rule_suboption_set,
)
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue, Rule
from suricata_check.utils.regex import get_regex_provider

_regex_provider = get_regex_provider()

_S400_REGEX = _regex_provider.compile(r"""^"[A-Z0-9 ]+ [A-Z0-9_]+ .*"$""")
_MALWARE_REGEX = _regex_provider.compile(r"^.*(malware).*$", _regex_provider.IGNORECASE)
_S401_REGEX = _regex_provider.compile(r"""^".* [a-zA-Z0-9]+/[a-zA-Z0-9]+ .*"$""")
_VAGUE_KEYWORDS = ("possible", "unknown")
_S402_REGEX = _regex_provider.compile(
    r"^.*({}).*$".format("|".join(_VAGUE_KEYWORDS)), _regex_provider.IGNORECASE
)
_UNDESIRABLE_DATE_REGEXES = (
    _regex_provider.compile(r"^.*(\d{4}/\d{2}/\d{2}).*$", _regex_provider.IGNORECASE),
    _regex_provider.compile(r"^.*(\d{4}-[2-9]\d-\d{2}).*$", _regex_provider.IGNORECASE),
)  # Desirable format is ISO (YYYY-MM-DD)
_S404_REGEX = _regex_provider.compile(
    r"^.*(C2|C&C|Command and Control|Command & Control).*$", _regex_provider.IGNORECASE
)
_S405_REGEX = _regex_provider.compile(
    r"^.*(Go|MSIL|ELF64|MSIL|JS|Win32|DOS|Amiga|C64|Plan9).*$",
    _regex_provider.IGNORECASE,
)
_S406_REGEX = _regex_provider.compile(
    r"^.*((\w+\.)+[a-z]{2,}).*$",
    _regex_provider.IGNORECASE,
)
_S407_REGEX = _regex_provider.compile(
    r"^.*((\w+\[\.\])+[a-z]{2,}).*$",
    _regex_provider.IGNORECASE,
)
_S408_REGEX = _regex_provider.compile(
    r"^.*((\w+\. )+[a-z]{2,}).*$",
    _regex_provider.IGNORECASE,
)

_logger = logging.getLogger(__name__)


class MsgChecker(CheckerInterface):
    """The `MsgChecker` contains several checks based for the Msg option in Suricata rules.

    Codes S400-S410 report on non-standard `msg` fields.
    """

    codes = {
        "S400": {"severity": logging.INFO},
        "S401": {"severity": logging.INFO},
        "S402": {"severity": logging.INFO},
        "S403": {"severity": logging.INFO},
        "S404": {"severity": logging.INFO},
        "S405": {"severity": logging.INFO},
        "S406": {"severity": logging.WARNING},
        "S407": {"severity": logging.INFO},
        "S408": {"severity": logging.INFO},
        "S409": {"severity": logging.INFO},
    }

    def _check_rule(  # noqa: C901
        self: "MsgChecker",
        rule: Rule,
    ) -> ISSUES_TYPE:
        issues: ISSUES_TYPE = []

        if is_rule_option_set(rule, "msg") and not is_rule_option_equal_to_regex(
            rule, "msg", _S400_REGEX
        ):
            issues.append(
                Issue(
                    code="S400",
                    message="""\
The rule has a non-standard format for the msg field.
Consider changing the msg field to `RULESET CATEGORY Description`.\
""",
                ),
            )

        if (
            is_rule_option_set(rule, "msg")
            and self.__desribes_malware(rule)
            and not is_rule_option_equal_to_regex(rule, "msg", _S401_REGEX)
        ):
            issues.append(
                Issue(
                    code="S401",
                    message="""\
The rule describes malware but does not specify the paltform or malware family in the msg field.
Consider changing the msg field to include `Platform/malfamily`.\
""",
                ),
            )

        if not (
            is_rule_option_set(rule, "noalert")
            or is_rule_suboption_set(rule, "flowbits", "noalert")
        ) and is_rule_option_equal_to_regex(rule, "msg", _S402_REGEX):
            issues.append(
                Issue(
                    code="S402",
                    message="""\
The rule uses vague keywords such as possible or unknown in the msg field.
Consider rephrasing to provide a more clear message for interpreting generated alerts.\
""",
                ),
            )

        for regex in _UNDESIRABLE_DATE_REGEXES:
            if is_rule_option_equal_to_regex(rule, "msg", regex):
                issues.append(
                    Issue(
                        code="S403",
                        message="""\
The rule uses a non-ISO date in the msg field.
Consider reformatting the date to ISO format (YYYY-MM-DD).\
""",
                    ),
                )
                break

        if is_rule_option_equal_to_regex(rule, "msg", _S404_REGEX):
            issues.append(
                Issue(
                    code="S404",
                    message="""\
The rule uses a different way of writing CnC (Command & Control) in the msg field.
Consider writing CnC instead.\
""",
                ),
            )

        if self.__desribes_malware(rule) and not is_rule_option_equal_to_regex(
            rule, "msg", _S405_REGEX
        ):
            issues.append(
                Issue(
                    code="S405",
                    message="""\
The rule likely detects malware but does not specify the file type in the msg field.
Consider specifying a file type such as `DOS` or `ELF64`.\
""",
                ),
            )

        if is_rule_option_equal_to_regex(rule, "msg", _S406_REGEX):
            issues.append(
                Issue(
                    code="S406",
                    message="""\
The rule specifies a domain name without escaping the label seperators.
Consider escaping the domain names by putting a space before the dot like `foo .bar` to prevent information leaks.\
""",
                ),
            )

        if is_rule_option_equal_to_regex(rule, "msg", _S407_REGEX):
            issues.append(
                Issue(
                    code="S407",
                    message="""\
The rule specifies a domain name and escapes it in a non-standard way in the msg field.
Consider escaping the domain names by putting a space before the dot like `foo .bar`.\
""",
                ),
            )

        if is_rule_option_equal_to_regex(rule, "msg", _S408_REGEX):
            issues.append(
                Issue(
                    code="S408",
                    message="""\
The rule specifies a domain name and escapes it in a non-standard way in the msg field.
Consider escaping the domain names by putting a space before the dot like `foo .bar`.\
""",
                ),
            )

        # Note that all characters under 128 are ASCII
        if is_rule_option_set(rule, "msg") and any(
            ord(c) > 128  # noqa: PLR2004
            for c in get_rule_option(rule, "msg")  # type: ignore reportOptionalIterable
        ):
            issues.append(
                Issue(
                    code="S409",
                    message="""\
The rule uses non-ASCII characters in the msg field.
Consider removing non-ASCII characters.\
""",
                ),
            )

        return issues

    @staticmethod
    def __desribes_malware(rule: Rule) -> bool:
        if is_rule_suboption_set(rule, "metadata", "malware_family"):
            return True

        if is_rule_option_equal_to_regex(rule, "msg", _MALWARE_REGEX):
            return True

        _logger.debug("Rule does not describe malware: %s", rule["raw"])

        return False



"""The `suricata_check.checkers.interface` modules contains the interfaces implemented by checkers.

Implementation of the `CheckerInterface` is neccessary for checker auto-discovery.
"""

from suricata_check.checkers.interface.checker import CheckerInterface
from suricata_check.checkers.interface.dummy import DummyChecker

__all__ = ["CheckerInterface", "DummyChecker"]



"""`StateChecker`."""

import logging

from suricata_check.checkers.interface import CheckerInterface
from suricata_check.utils.checker import (
    get_rule_suboptions,
    is_rule_option_always_equal_to_regex,
    is_rule_option_always_put_before,
    is_rule_option_set,
    is_rule_suboption_set,
)
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue, Rule
from suricata_check.utils.regex import get_regex_provider

_regex_provider = get_regex_provider()

_S510_REGEX = _regex_provider.compile(
    r"^((set|isset|toggle|unset|isnotset),[A-Z]+\.\w+)|(noalert)$",
    _regex_provider.IGNORECASE,
)

_S520_REGEX = _regex_provider.compile(
    r"^((set|isset|toggle|unset|isnotset),[A-Z]+\.\w+)|(noalert),.*$",
    _regex_provider.IGNORECASE,
)

_logger = logging.getLogger(__name__)


class StateChecker(CheckerInterface):
    """The `StateChecker` contains several checks for Suricata options relating to the connection or detection state.

    Codes S500-S510 report on non-standard usages of `flow`
    Codes S510-S520 report on non-standard usages of `flowbits`
    Codes S520-S530 report on non-standard usages of `xbits`
    """

    codes = {
        "S500": {"severity": logging.INFO},
        "S501": {"severity": logging.INFO},
        "S510": {"severity": logging.INFO},
        "S511": {"severity": logging.INFO},
        "S520": {"severity": logging.INFO},
        "S521": {"severity": logging.INFO},
        "S522": {"severity": logging.INFO},
    }

    def _check_rule(
        self: "StateChecker",
        rule: Rule,
    ) -> ISSUES_TYPE:
        issues: ISSUES_TYPE = []

        if is_rule_option_set(rule, "flow") and not is_rule_option_always_put_before(
            rule,
            "established",
            ["to_server", "to_client", "from_server", "from_client"],
            sequence=[suboption[0] for suboption in get_rule_suboptions(rule, "flow")],
        ):
            issues.append(
                Issue(
                    code="S500",
                    message="""\
The rule specifies the connection state after the connection direction in the `flow` option.
Consider specifying the connection state first like `established,to_server`.\
""",
                ),
            )

        if is_rule_suboption_set(rule, "flow", "from_client") or is_rule_suboption_set(
            rule, "flow", "from_server"
        ):
            issues.append(
                Issue(
                    code="S501",
                    message="""\
The rule has set `from_client` or `from_server` as a `flow` option.
Consider using `to_client` or `to_server` instead.\
""",
                ),
            )

        if is_rule_option_always_equal_to_regex(rule, "flowbits", _S510_REGEX) is False:
            issues.append(
                Issue(
                    code="S510",
                    message="""\
The rule sets flowbits with a non-standard name.
Consider using `RULESET.description` as name for the flowbit.\
""",
                ),
            )

        if (
            (
                is_rule_suboption_set(rule, "flowbits", "set")
                or is_rule_suboption_set(rule, "flowbits", "unset")
            )
            and not (
                is_rule_suboption_set(rule, "flowbits", "isset")
                or is_rule_suboption_set(rule, "flowbits", "isnotset")
            )
            and not (
                is_rule_option_set(rule, "noalert")
                or is_rule_suboption_set(rule, "flowbits", "noalert")
            )
        ):
            issues.append(
                Issue(
                    code="S511",
                    message="""\
The rule (un)sets a flowbit but does not use the noalert option.
Consider using the noalert option to prevent unnecessary alerts.\
""",
                ),
            )

        if is_rule_option_always_equal_to_regex(rule, "xbits", _S520_REGEX) is False:
            issues.append(
                Issue(
                    code="S520",
                    message="""\
The rule sets xbits with a non-standard name.
Consider using `RULESET.description` as name for the xbit.\
""",
                ),
            )

        if (
            (
                is_rule_suboption_set(rule, "xbits", "set")
                or is_rule_suboption_set(rule, "xbits", "unset")
            )
            and not is_rule_option_set(rule, "noalert")
            and not (
                is_rule_suboption_set(rule, "xbits", "isset")
                or is_rule_suboption_set(rule, "xbits", "isnotset")
            )
        ):
            issues.append(
                Issue(
                    code="S521",
                    message="""\
The rule (un)sets a xbit but does not use the noalert option.
Consider using the noalert option to prevent unnecessary alerts.\
""",
                ),
            )

        if (is_rule_suboption_set(rule, "xbits", "set")) and not is_rule_suboption_set(
            rule, "xbits", "expire"
        ):
            issues.append(
                Issue(
                    code="S522",
                    message="""\
The rule sets a xbit but does not explcitly set the expire option.
Consider setting the expire option to indicate for how long the xbit remains relevant.\
""",
                ),
            )

        return issues



"""`OverallChecker`."""

import logging

from suricata_check.checkers.interface import CheckerInterface
from suricata_check.utils.checker import (
    count_rule_options,
    get_all_variable_groups,
    get_rule_option,
    get_rule_sticky_buffer_naming,
    is_rule_option_equal_to,
    is_rule_option_equal_to_regex,
    is_rule_option_one_of,
    is_rule_option_set,
)
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue, Rule
from suricata_check.utils.regex import (
    ALL_VARIABLES,
    CLASSTYPES,
    get_regex_provider,
)

_regex_provider = get_regex_provider()


# Regular expressions are placed here such that they are compiled only once.
# This has a significant impact on the performance.
_REGEX_S002a = _regex_provider.compile(
    r"^.*(EXPLOIT|CVE).*$",
    _regex_provider.IGNORECASE,
)
_REGEX_S002b = _regex_provider.compile(
    r"^.*(Internal|Inbound|Outbound).*$",
    _regex_provider.IGNORECASE,
)
_REGEX_S030 = _regex_provider.compile(r"^[a-z\-]+$")
_REGEX_S031 = _regex_provider.compile(r"^[^\|]*\|[^\|]*[A-Z]+[^\|]*\|[^\|]*$")


class OverallChecker(CheckerInterface):
    """The `OverallChecker` contains several the most basic checks for Suricata rules.

    Codes S000-S009 report on issues with the direction of the rule.

    Codes S010-S019 report on issues pertaining to the usage of non-standard options.

    Codes S020-S029 report on issues pertaining to the non-usage of recommended options.

    Codes S020-S029 report on issues pertaining to the non-usage of recommended options.

    Codes S031-S039 report on issues pertaining to the inappropriate usage of options.
    """

    codes = {
        "S000": {"severity": logging.INFO},
        "S001": {"severity": logging.INFO},
        "S002": {"severity": logging.INFO},
        "S010": {"severity": logging.INFO},
        "S011": {"severity": logging.INFO},
        "S012": {"severity": logging.INFO},
        "S013": {"severity": logging.INFO},
        "S014": {"severity": logging.INFO},
        "S020": {"severity": logging.INFO},
        "S021": {"severity": logging.INFO},
        "S030": {"severity": logging.INFO},
        "S031": {"severity": logging.INFO},
    }

    def _check_rule(  # noqa: C901
        self: "OverallChecker",
        rule: Rule,
    ) -> ISSUES_TYPE:
        issues: ISSUES_TYPE = []

        if is_rule_option_equal_to(rule, "direction", "<->") or (
            is_rule_option_equal_to(rule, "source_addr", "any")
            and is_rule_option_equal_to(rule, "dest_addr", "any")
        ):
            issues.append(
                Issue(
                    code="S000",
                    message="""The rule did not specificy an inbound or outbound direction.
Consider constraining the rule to a specific direction such as INBOUND or OUTBOUND traffic.""",
                )
            )

        if is_rule_option_set(rule, "dns.query") and not is_rule_option_equal_to(
            rule,
            "dest_addr",
            "any",
        ):
            issues.append(
                Issue(
                    code="S001",
                    message="""The rule detects certain dns queries and has dest_addr not set to any \
causing the rule to be specific to either local or external resolvers.
Consider setting dest_addr to any.""",
                )
            )

        if (
            is_rule_option_equal_to_regex(rule, "msg", _REGEX_S002a)
            and not (
                is_rule_option_equal_to(rule, "source_addr", "any")
                and is_rule_option_equal_to(rule, "dest_addr", "any")
            )
            and not is_rule_option_equal_to_regex(rule, "msg", _REGEX_S002b)
        ):
            issues.append(
                Issue(
                    code="S002",
                    message="""The rule detects exploitation attempts in a constrained direction \
without specifying the direction in the rule msg. \
Consider setting `src_addr` and `dest_addr` to any to also account for lateral movement scenarios. \
Alternatively, you can specify the direction (i.e., `Internal` or `Inbound`) in the rule `msg`.""",
                )
            )

        # In the suricata style guide, this is mentioned as `packet_data`
        if is_rule_option_set(rule, "pkt_data"):
            issues.append(
                Issue(
                    code="S010",
                    message="""The rule uses the pkt_data option, \
which resets the inspection pointer resulting in confusing and disjoint logic.
Consider replacing the detection logic.""",
                )
            )

        if is_rule_option_set(rule, "priority"):
            issues.append(
                Issue(
                    code="S011",
                    message="""The rule uses priority option, which overrides operator tuning via classification.conf.
Consider removing the option.""",
                )
            )

        for sticky_buffer, modifier_alternative in get_rule_sticky_buffer_naming(rule):
            issues.append(
                Issue(
                    code="S012",
                    message=f"""The rule uses sticky buffer naming in the {sticky_buffer} option, which is complicated.
Consider using the {modifier_alternative} option instead.""",
                )
            )

        for variable_group in self.__get_invented_variable_groups(rule):
            issues.append(
                Issue(
                    code="S013",
                    message=f"""The rule uses a self-invented variable group ({variable_group}), \
which may be undefined in many environments.
Consider using the a standard variable group instead.""",
                )
            )

        if is_rule_option_set(rule, "classtype") and not is_rule_option_one_of(
            rule, "classtype", CLASSTYPES
        ):
            issues.append(
                Issue(
                    code="S014",
                    message=f"""The rule uses a self-invented classtype ({get_rule_option(rule, 'classtype')}), \
which may be undefined in many environments.
Consider using the a standard classtype instead.""",
                )
            )

        if not is_rule_option_set(rule, "content"):
            issues.append(
                Issue(
                    code="S020",
                    message="""The detection logic does not use the content option, \
which is can cause significant runtime overhead.
Consider adding a content match.""",
                )
            )

        if (
            not is_rule_option_set(rule, "fast_pattern")
            and count_rule_options(rule, "content") > 1
        ):
            issues.append(
                Issue(
                    code="S021",
                    message="""The rule has multiple content matches but does not use fast_pattern.
Consider assigning fast_pattern to the most unique content match.""",
                )
            )

        if is_rule_option_equal_to_regex(
            rule,
            "app-layer-protocol",
            _REGEX_S030,
        ):
            issues.append(
                Issue(
                    code="S030",
                    message="""The rule uses app-layer-protocol to assert the protocol.
Consider asserting this in the head instead using {} {} {} {} {} {} {}""".format(
                        get_rule_option(rule, "action"),
                        get_rule_option(rule, "app-layer-protocol"),
                        get_rule_option(rule, "source_addr"),
                        get_rule_option(rule, "source_port"),
                        get_rule_option(rule, "direction"),
                        get_rule_option(rule, "dest_addr"),
                        get_rule_option(rule, "dest_port"),
                    ),
                )
            )

        if is_rule_option_equal_to_regex(
            rule,
            "content",
            _REGEX_S031,
        ):
            issues.append(
                Issue(
                    code="S031",
                    message="The rule uses uppercase A-F in a hex content match.\nConsider using lowercase a-f instead.",
                )
            )

        return issues

    @staticmethod
    def __get_invented_variable_groups(
        rule: Rule,
    ) -> list[str]:
        variable_groups = get_all_variable_groups(rule)

        invented_variable_groups = []

        for variable_group in variable_groups:
            if variable_group not in ALL_VARIABLES:
                invented_variable_groups.append(variable_group)

        return invented_variable_groups
